/** \file appconfig.c	Configuration module.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup config
 *
 * The appconfig module manages the gtksql configuration files.  At present
 * there are three different types of config files:
 * - The resource file ${HOME}/.gtksqlrc
 * - The config file ${HOME}/.gtksql/gtksql.conf
 * - Script files.
 *
 * The resource file is in the standard gtk resource format.
 * The config file and the script files are both LUA scripts. gtksql.conf is loaded every time the application runs.
 * Other scripts are loaded either from the command line, or from menus inside the application.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2003  Lionel ULMER, Darryl Luff
 * $Id: appconfig.c,v 1.33 2003/09/01 16:00:02 djluff Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Some config items are stored in a gtk resource file int he standard gtk format.
 * But gtk doesnt seem to allow you to put arbitrary things in there. So the 
 * remainder of things are stored in a simple text config file.
 *
 * This module handles the differences transparently, so the application doesn't need
 * to know about these seperate files.
 */

#include "../config.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "appconfig.h"
#include "buf.h"
#include "dbapi.h"
#include "luaif.h"

/** Directory name for GtkSQL config files.	*/
#define CONFIG_DIRNAME ".gtksql"

/** Name of the main configuration file.	*/
#define CONFIG_FILENAME	"gtksql.conf"

#define CONTACT	"http://gtksql.sourceforge.net/"

/** Usage string to be displayed to the user.	*/
#ifdef USE_LUA
#define USAGE	"gtksql [-h] [-f scriptfile] [-s \"script\"]\n\n" \
				"\t-h              Show usage help\n"	\
				"\t-f scriptfile   Run the script in the named file\n"	\
				"\t-s \"script\"     Run the script given on the command line\n"

#define ARGSPEC	"hf:s:"
#else
#define USAGE	"gtksql [-h] \n\n" \
				"\t-h              Show usage help\n"

#define ARGSPEC	"h"
#endif

/** The path to the configuration directory.	*/
static char *cfgDirname = NULL;

/** The last accessed directory.	*/
static char *currentDir = NULL;


/** Show a usage string	*/
void cfg_show_banner(void)
{
	printf("%s Version %s\n%s\n", PACKAGE, VERSION, CONTACT);
	printf("Support for:\n");
	printf("\t      MySQL:\t");
#ifdef USE_MYSQL
	printf("Yes\n");
#else
	printf("No\n");
#endif
	printf("\t PostgreSQL:\t");
#ifdef USE_POSTGRESQL
	printf("Yes\n");
#else
	printf("No\n");
#endif

	printf("\tLUA Scripts:\t");
#ifdef USE_LUA
	printf("Yes\n");
#else
	printf("No\n");
#endif

	printf("\t      XPath:\t");
#ifdef USE_XPATH
	printf("Yes\n");
#else
	printf("No\n");
#endif
}

/** Show a usage string	*/
void cfg_show_usage(void)
{
	cfg_show_banner();
	printf("\nUSAGE: %s\n", USAGE);
}

/** Set the default configuration values.	*/
static void cfg_set_defaults(void)
{
	/* Dont have to do anything. Defaults hard-coded. */
}

/** Parses the command line.
 * @param	argc	As passed to main()
 * @param	argv	As passed to main()
 * @return	Always returns zero.	*/
int cfg_parse_cmdline(int argc, char *argv[])
{
	char c;
	while ((c = getopt(argc, argv, ARGSPEC)) != -1) {
		switch (c) {
		case 'c':
			/* Connect to a URL. */
			db_connect_url(optarg);
			break;
#ifdef USE_LUA
		case 'f':
			/* Run a script file. */
			luaif_runfile(optarg);
			break;
		case 's':
			/* Run a script directly. */
			luaif_runstring(optarg);
			break;
#endif
		case 'h':
			cfg_show_usage();
			break;
		}
	}
	return 0;
}

/** Initialise the configuration.
 * Performs these steps:
 *  - Sets default values.
 */
void cfg_init(void)
{
	cfg_set_defaults();
}

/** Get the name of the configuration directory.
 * This is normally ($HOME/.gtksql).
 * \return	The name of the configuration directory.	*/
char *cfg_get_config_dir(void)
{
	char *homedir;
	int len;
	homedir = getenv("HOME");

	if (!homedir) {
		printf("Could not resolve home directory");
		return NULL;
	}

	/* Set the current dir if it's still NULL. */
	if (currentDir == NULL) {
		cfg_set_current_dir(homedir);
	}

	len = strlen(homedir) + 1 + strlen(CONFIG_DIRNAME) + 1;
	if (cfgDirname)
		free(cfgDirname);

	cfgDirname = malloc(len);
	if (cfgDirname) {
		snprintf(cfgDirname, len, "%s/%s", homedir, CONFIG_DIRNAME);
		cfgDirname[len - 1] = '\0';
	}
	printf("Using config directory %s", cfgDirname);

	if (chdir(cfgDirname) == -1) {	   /* couldnt change to it */
		if (errno == ENOENT) {		   /* didnt exist, try and create it */
			if (mkdir(cfgDirname, 0755) == -1) {
				perror("Could not create config directory");
				free(cfgDirname);
				cfgDirname = NULL;
				return NULL;
			}
		}
		else {
			perror("changing to config directory");
			free(cfgDirname);
			cfgDirname = NULL;
			return NULL;
		}
	}
	return cfgDirname;
}

/** Get the name of the current working directory.
 * \return	The name of the configuration directory.	*/
char *cfg_get_current_dir(void)
{
	if (!currentDir)
		currentDir = strdup(getenv("HOME"));

	if (!currentDir)
		cfg_set_current_dir(".");

	return currentDir;
}

static ABuf *wsname = NULL;
/** Get the default workspace name.
 * \return The default worlspace filename.	*/
char *cfg_get_default_wsname(void)
{
#define DEFAULT_WSNAME	"workspace"
	wsname = buf_strcpy(wsname, cfg_get_config_dir());
	wsname = buf_strcat(wsname, "/");
	wsname = buf_strcat(wsname, DEFAULT_WSNAME);

	return wsname->b_dat;
}

/** Get the name of the config file.	
 * @return	The file name.	*/
/*
static char *cfg_get_filename(void)
{
	gchar *dir;
	int len;
	if (!cfgname) {
		dir = cfg_get_config_dir();

		len = strlen(dir) + strlen(CONFIG_FILENAME) + 2;
		cfgname = g_malloc(len);
		if (!cfgname)
			return NULL;

		snprintf(cfgname, len, "%s/%s", dir, CONFIG_FILENAME);
	}
	return cfgname;
}
*/

/** Displays a text version of the current configuration.	*/
/*static void cfg_print(void)
{
	printf("currentWS.changed = %d", ws_get_changed());
	printf("currentWS.dbname = %s\n", ws_get_db_name());
	printf("currentWS.dbtype = %s\n", ws_get_db_type());
	printf("currentWS.filename = %s\n", ws_get_filename());
	printf("currentWS.host = %s\n", ws_get_host());
	printf("currentWS.user = %s\n", ws_get_user());
	printf("currentWS.pass = %s\n", ws_get_pass());
	printf("currentWS.connection = %p\n", ws_get_connection());
	printf("currentWS.toolbar_conf = %d", ws_get_toolbarconf());
	printf("currentWS.windowX = %d", ws_get_window_x());
	printf("currentWS.windowY = %d", ws_get_window_y());
	printf("currentWS.windowWidth = %d", ws_get_window_width());
	printf("currentWS.windowHeight = %d", ws_get_window_height());
}
*/
/** Set the current working directory.
 * \param dir The directory to use.	*/
void cfg_set_current_dir(const char *dir)
{
	if (currentDir != NULL) {
		free(currentDir);
		currentDir = NULL;
	}

	if (dir)
		currentDir = strdup(dir);
}
