/** \file buf.c General-purpose dynamic buffer.
 * \author Darryl Luff
 *
 * A buffer pointer is passed as the first argument to every function call.
 * All functions return a pointer to the buffer. Note that this may NOT be the 
 * same pointer that went in. For example:
 *
 * \code
 * ABuf *buffer = buf_new(4);
 * buffer = buf_cat(buffer, "123456");
 * \endcode
 *
 * Here, 'buffer's value after the call will be different to the 'buffer's 
 * value before the call, because the buffer is reallocated to fit in the new 
 * contents. So for any function call that can cause the buffer to grow, always 
 * save the return value into the buffer.
 * \par
 * Functions are all NULL-safe. If a null buffer is passed in, a new buffer will
 * be allocated and returned. So instead of the above, could just say:
 * \code
 * ABuf *buffer = buf_cpy(NULL, "123456");
 * \endcode
 * To minimise reallocating memory, it is a good idea to use buf_new to allocate 
 * a buffer with a reasonable size though.
 * \par
 * When finished with the buffer, call buf_free on it. buf_free always returns a
 * \c NULL buffer pointer. This is to let you do:
 * \code
 * buffer = buf_free(buffer);
 * \endcode
 * So the buffer pointer is set to NULL at the same time the memory is freed.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002-2003 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/** A dynamic buffer that adjusts it's size as things are written to it.	*/

#include <stdlib.h>
#include <string.h>
#include "buf.h"

/** Minimum size buffer allocated. Any buffer allocation will return a buffer 
 * that is at least this size, even \c buf_new(0).	*/
#define MINSIZE	10

/** Returns the size of buffer to allocate, given the requested size.
 * \param n	The requested size.
 * \return	Either the requested size or \ref MINSIZE, whichever is smaller.	*/
int allocsize(int n) {
	if (n < MINSIZE)
		return MINSIZE;
	else
		return n;
}

/** Allocate a new ABuf. \ref allocsize() is used to calculate the buffer size,
 * so the returned buffer may be larger than the requested size.
 * \return	A pointer to the new buffer.
 * \see	MINSIZE	*/
ABuf *buf_new(unsigned int len)
{
	ABuf *b = calloc(1, sizeof(ABuf));
	if (b) {
		b->b_len = allocsize(len);
		b->b_dat = calloc(1, b->b_len);
	}
	return b;
}

/** Free the passed buffer.
 * \return NULL pointer so you can use
 * \code
 * 'buf = buf_free(buf);
 * \endcode
 * to free the buffer and clear the variable in one line.	*/
ABuf *buf_free(ABuf * buf)
{
	if (buf) {
		if (buf->b_dat)
			free(buf->b_dat);

		free(buf);
	}
	return NULL;
}

/** Make sure that the passed buffer is large enough. If it is not at least
 * 'len' bytes long, a new buffer is allocated. Returns a pointer to either
 * the original or the new buffer. Always use:
 * \code
 * buf = buf_check(buf, len);
 * \endcode
 * with this function, and never rely on the returned pointer being the same as
 * the passed pointer.
 * \param	buf	The original pointer. If \c NULL, a new buffer will be allocated.
 * \param	len	The requested buffer size.
 * \return	Either the original buffer if it was large enough, or a new larger buffer.	*/
ABuf *buf_check(ABuf * buf, unsigned int len)
{
	if (!buf)
		return buf_new(len);

	if (buf->b_len < len) {
		ABuf *b = buf_new(len);
		memmove(b->b_dat, buf->b_dat, buf->b_len);
		buf_free(buf);
		return b;
	}
	return buf;
}

/** Appends the string 's' to the buffer 'buf'. The buffer is reallocated if
 * necessary.
 * \code
 * buf = buf_strcat(buf, "Hello");
 * \endcode
 * \param	buf	The original buffer
 * \param	s	The string to append
 * \return	The new buffer	*/
ABuf *buf_strcat(ABuf * buf, char *s)
{
	ABuf *b = NULL;
	if (! buf)
		b = buf_new(strlen(s)+1);
	else
		b = buf_check(buf, strlen(buf->b_dat) + strlen(s) + 1);

	if (b) {
		strncat(b->b_dat, s, b->b_len);
		b = buf_terminate(b);
	}

	return b;
}

/** Copies the string 's' into the buffer. The buffer is reallocated 
 * if required.
 * \code
 * buf = buf_strcpy(buf, "Hello");
 * \endcode
 * \param	buf	The original buffer
 * \param	s	The string to copy
 * \return	The new buffer.	*/
ABuf *buf_strcpy(ABuf * buf, char *s)
{
	ABuf *b = NULL;
	int len;

	if (! s)
		return buf;

	len = strlen(s) + 1;
	if (! buf)
		b = buf_new(len);
	else
		b = buf_check(buf, len);

	if ((b) && (b->b_dat)) {
		if ((s) && (strlen(s) > 0)) {
			strncpy(b->b_dat, s, b->b_len);
			buf_terminate(b);
		}
		else
			*(b->b_dat) = '\0';
	}
	return b;
}

/** Sets the last non-NULL byte of the string to NULL, effectively shortening the string by one
 * character.
 * \param	buf	The original buffer.
 * \return	The buffer.	*/
ABuf *buf_str_shrinkbyone(ABuf * buf) {
	if ((buf == NULL) || (buf->b_dat == NULL) || (strlen(buf->b_dat) < 1))
		return buf;

	buf->b_dat[strlen(buf->b_dat)-1] = '\0';
	return buf;
}

/** Sets the last byte of the buffer to a null byte, overwriting whatever's
 * in it.
 * \param	buf	The original buffer.
 * \return	The buffer.	*/
ABuf *buf_terminate(ABuf * buf)
{
	if (buf) {
		if ((buf->b_dat) && (buf->b_len)) {
			buf->b_dat[buf->b_len - 1] = '\0';
		}
	}
	return buf;
}
