/** \file callbacks.c	Callbacks for Glade-generated interface.
 * \ingroup gui
 * \author Darryl Luff	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002-2003 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#include <string.h>

#include "buf.h"
#include "db_connect_dlg.h"
#include "dialogs.h"
#include "gui.h"
#include "guiapi.h"
#include "interface.h"
#include "luaif.h"
#include "queries.h"
#include "query_import_dlg.h"
#include "query_save_dlg.h"
#include "support.h"
#include "tables.h"
#include "workspace.h"

extern void disconnect_callback(void *data);

void on_mainWindow_destroy(GtkObject * object, gpointer user_data)
{
	DBConnection *conn = gui_get_connection();
	if (conn != NULL)
		disconnect_callback(user_data);

	gtk_main_quit();
}

/** Load a query from disk.	*/
void on_queryLoad_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	query_add();
	query_import();
}

/** Import a query from disk.	*/
void on_queryImport_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	query_import();
}

/** 'Save' query action function.	*/
void on_querySave_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	query_save(NULL);
}

/** Save a query to disk under a new name.	*/
void on_querySaveAs_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	query_save_as(NULL);
}

void on_resultExport_activate(GtkMenuItem * menuitem, gpointer user_data)
{

}

void on_quit_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	if (gui_connection_ok())
		disconnect_callback(user_data);

	gtk_main_quit();
}


/** Connect to database.
 * \param	menuitem	The widget sending the signal.
 * \param	user_data	Ignored.	*/
void on_databaseConnect_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	db_connect_dialog_show();
}


/** Disconnect action function.	*/
void on_databaseDisconnect_activate(GtkMenuItem * menuitem,
									gpointer user_data)
{
/** Template for disconnection message.	*/
#define SURE_S	" Are you sure you want \n to disconnect from database %s ? "
	GtkWidget *dialog;
	DBConnection *conn = gui_get_connection();
	ABuf *buf = buf_new(strlen(SURE_S) + strlen(conn->DBget_db_name(conn)));
	if (!buf) {
		g_warning("disconnect() could not allocate memory for a buffer\n");
		return;
	}

	if (conn == NULL) {
		display_error("Internal error",
					  "Internal error : trying to disconnect "
					  "with no current connection");
		return;
	}

	snprintf(buf->b_dat, buf->b_len, SURE_S, conn->DBget_db_name(conn));

	dialog =
		create_standard_dialog("Disconnect", buf->b_dat, 2, NULL, "Ok",
							   disconnect_callback, "Cancel", NULL, NULL);
	gtk_widget_show(dialog);
	buf_free(buf);
}

/** 'Send' query action function.	*/
void on_querySend_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	gui_send_query();
}

/** 'New' query action function.	*/
void on_queryNew_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	query_add();
}

/** Execute the clearQuery function.
 * \param menuitem	The sending widget.
 * \param user_data	Unused.	*/
void on_queryClear_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	query_clear();
}

/** 'Close' query action function.	*/
void on_queryClose_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	int nb_query;

	nb_query = query_close();

	if (nb_query == 0) {
		update_button_sensitivity();
	}
}

/** 'Refresh' action function.	*/
void on_tableRefresh_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	tables_connection_closed();
	tables_new_connection();
}

void on_configOptions_activate(GtkMenuItem * menuitem, gpointer user_data)
{

}

void on_helpAbout_activate(GtkMenuItem * menuitem, gpointer user_data)
{
/** About dialog text.	*/
#define VERSION_S	"\n " APPNAME " Version " VERSION "\n "	\
		"Configured " CONFDATE "\n"
	ABuf *buf = NULL;
	GtkLabel *lbl;

	if (aboutDialog == NULL) {
		aboutDialog = create_aboutDlg();
		buf = buf_strcpy(NULL, "");

#ifdef USE_POSTGRESQL
		buf = buf_strcat(buf, " PostgreSQL driver Installed \n");
#endif
#ifdef USE_MYSQL
		buf = buf_strcat(buf, " MySQL driver installed \n");
#endif
#ifdef USE_XPATH
		buf = buf_strcat(buf, " XPath driver installed \n");
#endif

		lbl =
			GTK_LABEL(lookup_widget
					  (GTK_WIDGET(aboutDialog), "aboutVersionLabel"));
		if (lbl != NULL)
			gtk_label_set_text(GTK_LABEL(lbl), VERSION_S);

		lbl =
			GTK_LABEL(lookup_widget
					  (GTK_WIDGET(aboutDialog), "aboutDriversLabel"));
		if (lbl != NULL)
			gtk_label_set_text(GTK_LABEL(lbl), buf->b_dat);

		buf_free(buf);
	}
	gtk_widget_show(aboutDialog);
}

void on_fileQuit_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	if (gui_connection_ok())
		disconnect_callback(user_data);

	gtk_main_quit();
}

/** OK callback for save dialog.
 * \param button	The OK button.
 * \param user_data	Unused.	*/
void on_querySaveDlgOkBtn_clicked(GtkButton * button, gpointer user_data)
{
	char *filename;
	if (querySaveData.query == NULL)
		querySaveData.query = query_get_current();

	if (querySaveData.qdata == NULL)
		querySaveData.qdata = query_get_querydata(querySaveData.query);

	/* We get the file name */
	filename =
		gtk_file_selection_get_filename(GTK_FILE_SELECTION(querySaveDialog));
	querydata_set_filename(querySaveData.qdata, filename);
	query_save_to_file(querySaveData.query, filename);
	query_save_dialog_hide();
}

/** Cancel callback used by the Import and Save dialog.
 * \param button	The cancel button.
 * \param user_data	Unused.	*/
void on_querySaveDlgCancelBtn_clicked(GtkButton * button, gpointer user_data)
{
	query_save_dialog_hide();
}

/** OK callback for import dialog.
 * \param button	The OK button.
 * \param user_data	Unused.	*/
void on_queryImportDlgOkBtn_clicked(GtkButton * button, gpointer user_data)
{
	char *filename, *raw_filename;
	ABuf *buf = buf_new(256);
	FILE *fp;
	GtkWidget *cq = query_get_current();
	QueryData *qd = gtk_object_get_data(GTK_OBJECT(cq), QDATA_KEY);

	query_clear();

	/* We get the file name */
	filename =
		gtk_file_selection_get_filename(GTK_FILE_SELECTION
										(queryImportDialog));
	raw_filename =
		gtk_entry_get_text(GTK_ENTRY(queryImportDialog->selection_entry));

	gtk_text_freeze(GTK_TEXT(cq));
	fp = fopen(filename, "r");
	if (fp != NULL) {
		while (fgets(buf->b_dat, buf->b_len, fp) != NULL)
			gtk_text_insert(GTK_TEXT(cq), NULL, NULL, NULL, buf->b_dat, -1);

		fclose(fp);
		querydata_set_filename(qd, filename);
		querydata_set_unchanged(qd);
		gui_set_query_name(filename);
	}
	gtk_text_thaw(GTK_TEXT(cq));
	buf_free(buf);
	query_import_dialog_hide();
}

/** OK callback for import dialog.
 * \param button	The Cancel button.
 * \param user_data	Unused.	*/
void on_queryImportDlgCancelBtn_clicked(GtkButton * button,
										gpointer user_data)
{
	query_import_dialog_hide();
}

void on_aboutOKBtn_clicked(GtkButton * button, gpointer user_data)
{
	gtk_widget_hide(aboutDialog);
}

void on_scriptLoadDlgOKBtn_clicked(GtkButton * button, gpointer user_data)
{
#ifdef USE_LUA
	char *filename;
	gtk_widget_hide(GTK_WIDGET(scriptOpenDialog));
		
	/* We get the file name */
	filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(scriptOpenDialog));
	luaif_runfile(filename);
#endif
}

void on_scriptLoadDlgCancelBtn_clicked(GtkButton * button, gpointer user_data)
{
	gtk_widget_hide(GTK_WIDGET(scriptOpenDialog));
}

void on_scriptSaveDlgOKBtn_clicked(GtkButton *button, gpointer user_data)
{
#ifdef USE_LUA
	char *filename;
	gtk_widget_hide(GTK_WIDGET(scriptSaveDialog));
		
	/* We get the file name */
	filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(scriptSaveDialog));
	ws_save_file(filename);
#endif
}

void on_scriptSaveDlgCancelBtn_clicked(GtkButton *button, gpointer user_data)
{
	gtk_widget_hide(GTK_WIDGET(scriptSaveDialog));
}

void on_scriptSaveAs_activate(GtkMenuItem *menuitem, gpointer user_data)
{
	if (scriptSaveDialog == NULL) {
		scriptSaveDialog = create_scriptSaveDlg();

		if (scriptSaveDialog == NULL) {
			g_warning("Could not initialise Script Save Dialog\n");
			return;
		}
		else {
			gtk_file_selection_complete(GTK_FILE_SELECTION(scriptSaveDialog), "*.lua");
		}
	}

	gtk_widget_show(GTK_WIDGET(scriptSaveDialog));
}

void on_fileScriptLoad_activate(GtkMenuItem *menuitem, gpointer user_data)
{
	if (scriptOpenDialog == NULL) {
		scriptOpenDialog = create_scriptOpenDlg();

		if (scriptOpenDialog == NULL) {
			g_warning("Could not initialise Script Open Dialog\n");
			return;
		}
		else {
			gtk_file_selection_complete(GTK_FILE_SELECTION(scriptOpenDialog), "*.lua");
		}
	}
	gtk_widget_show(GTK_WIDGET(scriptOpenDialog));
}
