/** \file db_connect_dlg.c	Database connect dialog.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup gui	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 1998-2003  Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of tdo_he License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#include <string.h>

#include "dialogs.h"
#include "guiapi.h"
#include "queries.h"
#include "register.h"
#include "tables.h"

/* Data keys	*/
#define	KEY_FIELDNAME	"fieldname"	/**< Field name data key. Used to associate a form field with a gtk_widget.	*/

extern void update_button_sensitivity(void);

/** Database connection dialog widget.	*/
static GtkWidget *dbConnectDialog = NULL;


/** Sets the value of a form field from a widget. If the widget has a
 * 'fieldname' data item set in it, the value of the text field is
 * read and store dinto the appropriate named field in the form.
 * \param w	The widget to query.
 * \param data	A pointer to the form.	*/
static void get_text_field(GtkWidget * w, gpointer data)
{
	Form *frm = (Form *) data;
	gchar *fname;
	gchar *s;

	if (!w) {
		return;
	}

	fname = (gchar *) gtk_object_get_data(GTK_OBJECT(w), KEY_FIELDNAME);
	if (fname) {
		s = gtk_editable_get_chars(GTK_EDITABLE(w), 0, -1);
		if (s) {
			frm_set_value_by_name(frm, (char *) fname, s);
			g_free(s);
			s = NULL;
		}
	}
}

/** Transfer fields from a the connection pane to a Form. Uses get_text_field()
 * on each GTK widget to extract the data.
 * \param frm	The form pointer.
 * \param pane	The connection pane.	*/
static void load_form_from_pane(Form * frm, GtkWidget * pane)
{
	gtk_container_foreach(GTK_CONTAINER(pane), get_text_field, frm);
}

/** Hide the active dialog.	*/
static void db_connect_dialog_hide(void)
{
	if (dbConnectDialog != NULL) {
		gtk_grab_remove(dbConnectDialog);
		gtk_widget_hide(GTK_WIDGET(dbConnectDialog));
	}
}

/** Reconnect using the login parameters in the login dialog.	*/
int db_connect_dialog_reconnect(GtkWidget * widget)
{
	GtkWidget *notebook = GTK_WIDGET(connectorTable[0]);
	GtkWidget *pane = NULL;
	GList *children = NULL;
	ABuf *errmsg = NULL;
	ABuf *err = NULL;

	int page;
	int i;
	DBConnection *conn = gui_get_connection();
	DBConnector *dbc = NULL;

	/* Make sure we're not currently connected. */
	if (conn != NULL) {
		display_error("Internal error",
					  "Internal error : connection should be closed");
		return CONNECT_ERROR;
	}

	/* First, we get the right connector function. */
	if (!notebook)
		return CONNECT_ERROR;

	page = gtk_notebook_current_page(GTK_NOTEBOOK(notebook));

	/* Because connector_table[0] = notebook. */
	page++;

	if (connectorTable[page] == NULL) {
		display_error("Internal error",
					  "Internal error : open dialog badly created");
		return CONNECT_ERROR;
	}

	/* The table is the first child of the notebook. Get it. */
	children = gtk_container_children(GTK_CONTAINER(notebook));
	if (!children)
		return CONNECT_ERROR;

	/* Skip ahead to the proper page. */
	for (i = 0; i < (page - 1); i++)
		children = children->next;

	pane = (GtkWidget *) (children->data);
	if (!pane)
		return CONNECT_ERROR;

	dbc = connectorTable[page];
	load_form_from_pane(dbc->db_frm, pane);
	errmsg = frm_check_fields(dbc->db_frm);
	if (errmsg == NULL) {
		conn = db_connect(dbc);
		if (conn == NULL) {
			g_warning("Couldn't finish connecting.\n");
			return CONNECT_ERROR;
		}
		if (conn->connectionStatus == CONNECT_OK) {
			gui_set_connection(conn);
		}
		return conn->connectionStatus;
	}
	else {
		err = buf_new(1024);
		err = buf_strcpy(err, "You did not complete some "
						 "required fields (");
		err = buf_strcat(err, errmsg->b_dat);
		err = buf_strcat(err, "). Please complete them and try " "again. \n");
		display_error("Connection error", err->b_dat);
		errmsg = buf_free(errmsg);
		err = buf_free(err);
	}
	return CONNECT_ERROR;
}

/** Sets a widget's text to the value of a form field. Looks up the 'fieldname'
 * data item from the widget to get the field name, then checks for that field
 * in the form. If it exists, its value is copied to the widget.
 * \param w	The widget to query.
 * \param data	A pointer to the form.	*/
static void set_text_field(GtkWidget * w, gpointer data)
{
	Form *frm = (Form *) data;
	gpointer *fname;
	gchar *s;
	gint pos = 0;

	if (!w) {
		return;
	}

	fname = gtk_object_get_data(GTK_OBJECT(w), KEY_FIELDNAME);
	if (fname) {
		s = frm_get_value_by_name(frm, (char *) fname);
		gtk_editable_delete_text(GTK_EDITABLE(w), 0, -1);
		if (s)
			gtk_editable_insert_text(GTK_EDITABLE(w), s, strlen(s), &pos);
	}
}

/** Transfer fields from a Form to the pane. For each field in the form,
 * locate a widget in the container with the same name as the field name,
 * and set it's text to the value of the field.
 * \param pane	The connection pane.
 * \param frm	The form pointer.	*/
static void load_pane_from_form(GtkWidget * pane, Form * frm)
{
	gtk_container_foreach(GTK_CONTAINER(pane), set_text_field, frm);
}

/** Execute the file open dialog for the file 'connect' function.
 * \param widget	The sending object.
 * \param data	???	*/
static void do_browse_file_dialog(GtkWidget * widget, gpointer data)
{
	file_open_dialog(data);
}

/** Creates a pane from a form.
 * \param	frm	The form data to use.	*/
static GtkWidget *make_form_pane(Form * frm)
{
	GtkWidget *table;
	GtkWidget *lbl;
	GtkWidget *txt;
	GtkWidget *btnLbl;
	GtkWidget *openBtn;
	FormField *fld;
	char *tiptext;
	char *helptext;
	char *fieldname;
	int i;
	int fieldcount;
	int ftype;
	ABuf *buf = NULL;
	int somerequired = 0;

	/* Creates the notebook pane and label */
	/* The connection pane */
	fieldcount = frm_get_field_count(frm);
	table = gtk_table_new(fieldcount + 1, 3, FALSE);
	gtk_container_border_width(GTK_CONTAINER(table), 6);
	for (i = 0; i < fieldcount; i++) {
		fld = frm_get_field(frm, i);
		fieldname = fld_get_name(fld);

		buf = buf_strcpy(buf, fieldname);
		if (fld_get_relevance(fld) == FR_REQUIRED) {
			buf = buf_strcat(buf, "*");
			somerequired++;
		}
		lbl = gtk_label_new(buf->b_dat);
		gtk_misc_set_padding(GTK_MISC(lbl), 4, 0);

		/* this doesnt seem to work? */
		gtk_label_set_justify(GTK_LABEL(lbl), GTK_JUSTIFY_RIGHT);
		gtk_widget_show(lbl);

		ftype = fld_get_type(fld);
		tiptext = fld_get_tip(fld);
		helptext = fld_get_help(fld);

		txt = gtk_entry_new();
		if (tiptext)
			gtk_tooltips_set_tip(gtk_tooltips_new(), txt, tiptext, helptext);

		gtk_widget_show(txt);
		switch (ftype) {
		case FT_TEXT:
			break;
		case FT_PWD:
			gtk_entry_set_visibility(GTK_ENTRY(txt), FALSE);
			break;
		case FT_FILENAME:
			openBtn = gtk_button_new();
			btnLbl = gtk_label_new("Browse...");
			gtk_misc_set_padding(GTK_MISC(btnLbl), 4, 0);
			gtk_widget_show(btnLbl);
			gtk_widget_show(openBtn);
			gtk_container_add(GTK_CONTAINER(openBtn), btnLbl);
			gtk_signal_connect(GTK_OBJECT(openBtn), "clicked",
							   GTK_SIGNAL_FUNC(do_browse_file_dialog), txt);

			gtk_table_attach(GTK_TABLE(table), openBtn, 2, 3, i, i + 1, 0, 0,
							 0, 0);
			break;
		}
		gtk_object_set_data(GTK_OBJECT(txt), KEY_FIELDNAME,
							strdup(frm_get_field_name(frm, i)));
		gtk_table_attach(GTK_TABLE(table), lbl, 0, 1, i, i + 1, GTK_SHRINK,
						 GTK_FILL | GTK_EXPAND, 0, 0);
		gtk_table_attach_defaults(GTK_TABLE(table), txt, 1, 2, i, i + 1);
	}
	if (somerequired > 0) {
		lbl = gtk_label_new("* = required fields");
		gtk_label_set_justify(GTK_LABEL(lbl), GTK_JUSTIFY_LEFT);
		gtk_widget_show(lbl);
		gtk_table_attach_defaults(GTK_TABLE(table), lbl, 1, 2, fieldcount,
								  fieldcount + 1);
	}
	buf_free(buf);
	load_pane_from_form(table, frm);
	return table;
}

/** Callback for 'Connect' dialog Cancel button.	*/
static void db_connect_dialog_cancel_callback(GtkWidget * widget,
											  gpointer data)
{
	GtkWidget *dialog = GTK_WIDGET(data);

	gtk_grab_remove(dialog);
	gtk_widget_hide(dialog);
}

/** Callback for 'Connect' dialog 'OK' button.
 * \param widget	The OK button.
 * \param data	The Connector table.	*/
static void db_connect_dialog_ok_callback(GtkWidget * widget, gpointer data)
{
	if (db_connect_dialog_reconnect(widget) == CONNECT_OK)
		db_connect_dialog_hide();
}

/** Create the connection dialog.	*/
void db_connect_dialog_create(void)
{
	int nb_db;
	if (dbConnectDialog != NULL)
		return;

	/* Count the number of databases supported */
	nb_db = 0;
	while (registerDB[nb_db] != NULL)
		nb_db++;

	if (nb_db == 0) {
		display_error("Notice",
					  "You do not have any database driver "
					  "installed. When GtkSQL was built it could not find any supported database.");
	}
	else {
		GtkWidget *ok_button, *cancel_button, *notebook;
		GtkWidget *lbl;
		int i;

		/* Create the connector table */
		connectorTable =
			(DBConnector **) malloc((nb_db + 2) * sizeof(DBConnector *));

		/* Creates the top-level dialog */
		dbConnectDialog = gtk_dialog_new();
		gtk_window_set_title(GTK_WINDOW(dbConnectDialog),
							 "Database connection");

		/* Buttons... */
		lbl = gtk_label_new("OK");
		gtk_misc_set_padding(GTK_MISC(lbl), 4, 0);
		gtk_widget_show(GTK_WIDGET(lbl));
		ok_button = gtk_button_new();
		gtk_container_add(GTK_CONTAINER(ok_button), lbl);
		gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
						   GTK_SIGNAL_FUNC(db_connect_dialog_ok_callback),
						   connectorTable);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dbConnectDialog)->action_area),
						   ok_button, FALSE, FALSE, 0);
		GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
		gtk_widget_grab_default(ok_button);
		gtk_widget_show(ok_button);

		lbl = gtk_label_new("Cancel");
		gtk_misc_set_padding(GTK_MISC(lbl), 4, 0);
		gtk_widget_show(GTK_WIDGET(lbl));
		cancel_button = gtk_button_new();
		gtk_container_add(GTK_CONTAINER(cancel_button), lbl);
		gtk_signal_connect(GTK_OBJECT(cancel_button), "clicked",
						   GTK_SIGNAL_FUNC(db_connect_dialog_cancel_callback),
						   dbConnectDialog);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dbConnectDialog)->action_area),
						   cancel_button, FALSE, FALSE, 0);
		gtk_widget_show(cancel_button);

		/* The fields notebook. */
		notebook = gtk_notebook_new();
		gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dbConnectDialog)->vbox),
						   notebook, FALSE, FALSE, 0);
		gtk_widget_show(notebook);

		/* Add the pages to the dialog and the callback functions */
		connectorTable[0] = (DBConnector *) notebook;
		for (i = 0; i < nb_db; i++) {
			GtkWidget *pane;
			Form *f;

			connectorTable[i + 1] = registerDB[i] (display_error);
			f = connectorTable[i + 1]->db_frm;
			if (f) {
				pane = make_form_pane(f);
				gtk_widget_show(pane);
				gtk_notebook_append_page(GTK_NOTEBOOK(notebook), pane,
										 gtk_label_new(frm_get_title(f)));
			}
		}
		connectorTable[i + 1] = NULL;
		update_button_sensitivity();
	}
}

/** Show the connect dialog.	*/
void db_connect_dialog_show(void)
{
	db_connect_dialog_create();
	gtk_window_position(GTK_WINDOW(dbConnectDialog), GTK_WIN_POS_MOUSE);
	gtk_window_set_policy(GTK_WINDOW(dbConnectDialog), FALSE, FALSE, FALSE);
	gtk_grab_add(dbConnectDialog);
	gtk_widget_show(dbConnectDialog);
}
