/** \addtogroup db Database Module
 * The functions in the database module provide the interface to the database drivers and database
 * connections. All calls from other parts of the program to the database system should be through 
 * the interface defined in this module.
 *
 *@{
 */
/** \file dbapi.c	Public database API.
 * \author Lionel ULMER
 * \author Darryl Luff
 */
/*@}*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2003 Lionel ULMER, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#include "dbapi.h"
#include "dburl.h"
#include "guiapi.h"

DBConnector **connectorTable;

/** Connect to the database defined in the current workspace.
 * \param url	The database URL to connect to.
 * \return The DBConnection.	*/
DBConnection *db_connect_url(char *url) {
	DBConnection *conn = NULL;
	DBConnector *dbc;
	Form *frm;
	int return_status;

	/* Get a connector for this URL	*/
	dbc = dburl_get_connector(url);
	if (dbc == NULL) {
		printf("Could not find a connector for the URL \"%s\"\n", url);
		return 0;
	}

	/* Set the login parameters	*/
	frm = dbc->db_frm;
	dbc->url_to_form(frm, url);

	/* Now, connect with it	*/
	return_status = dbc->db_connect_callback(dbc, &conn);
	if (conn != NULL)
		conn->connectionStatus = return_status;
	return conn;
}

/** Allocate a new generic DBConnector. All fields are NULL-filled.
 * @return	A pointer to a new DBConnector.	*/
DBConnector *dbconnector_new(void)
{
	return (DBConnector *) calloc(1, sizeof(DBConnector));
}

/** Connects to the database.
 * \param dbc	The connector to use.
 * \return A pointer to the new DBConnection structure, or NULL on error.	*/
DBConnection *db_connect(DBConnector *dbc) {
	DBConnection *conn = NULL;
	int return_status = dbc->db_connect_callback(dbc, &conn);
	if (conn != NULL)
		conn->connectionStatus = return_status;

	return conn;
}

/** Disconnects the current database.
 * \param conn	The connection to disconnect.
 * \param display_error	A pointer to an error 
 * function with this signature: void display_error(char *title, char *msg).
 * If not NULL, this function is called to display error messages if
 * required.	*/
void db_disconnect(DBConnection *conn, void(*display_error) (char *, char *)) {
	if (conn == NULL) {
		if (display_error != NULL)
			display_error("Internal error",
					  "Internal error : trying to disconnect "
					  "with no current connection");
		return;
	}
	conn->DBdisconnect(conn);
}
