/** \addtogroup db Database Module
 * The functions in the database module provide the interface to the database drivers and database
 * connections. All calls from other parts of the program to the database system should be through 
 * the interface defined in this module.
 *
 * \par
 * There are two main structures provided by the module:
 * - \ref DBConnector "DBConnector". This is an interface to a database driver. To connect to a database, you need to get a 
 * DBConnector for the specific database driver, and then use this connector to connect to the actual database.
 * - \ref DBConnection "DBConnection". When you connect to a database, you get back a DBConnection. This structure includes
 * pointers to functions to operate on the connection. These functions are used to send queries and access
 * the results of the query, among other things.
 *
 *@{
 */
/** \file dbapi.h	Public database API.
 * \author Lionel ULMER
 * \author Darryl Luff
 *
 * This file describes the API functions supported by the database module, and available for use by
 * other modules.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 1998-2003  Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of tdo_he License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _DBAPI_H
#define _DBAPI_H

#include "forms.h"

/** Query Status codes	*/
enum QueryStatus {
	DB_QUERY_ERROR = 0,
	DB_QUERY_EMPTY = 1,
	DB_QUERY_COMMAND_OK = 2,
	DB_QUERY_RECORDS_OK = 3,
	DB_QUERY_OTHER = 4
};

/** How to connect.	*/
enum ConnectStatus {
	CONNECT_DISCONNECTED = 0,
	CONNECT_ERROR,
	CONNECT_OK
};

/** DBConnector is the type used to refer to the _dbconnector structure.	*/
typedef struct _dbconnector DBConnector;

/** Auto-completion keyword list.	*/
typedef struct _keyword {
	char *key;				/**< Keyword text.	*/
	int type;				/**< Type of keyword.	*/
	struct _keyword *next;	/**< Pointer to next keyword in list.	*/
} Keyword;

/** Field record.	*/
typedef struct {
	char *name;		/**< Field name.	*/
	char *type;		/**< Field type string.	*/
	char *length;	/**< Field length.	*/
} DBFieldDef;

/** Table record.	*/
typedef struct {
	char *name;				/**< Table name.	*/
	int field_num;			/**< Number of fields.	*/
	DBFieldDef *field_def;	/**< Field definitions.	*/
} DBTableDef;

/** Per-connection database structure. This defines the interface the
 * application talks to the driver with. It also holds the keyword list
 * for the connection.	*/
struct db_conn {
	DBConnector *connector;	/**< A link back to the connector that created this connection.	*/
	int connectionStatus;	/**< \see ConnectStatus	*/

	/* Low-level SQL functions */
	/* Connection and status functions. */
	int (*DBdisconnect) (struct db_conn *);	/**< Disconnect from database.	*/

	/* Information functions. */
	char *(*DBget_error_message) (struct db_conn *);	/**< Return the last error message.	*/
	char *(*DBget_db_name) (struct db_conn *);			/**< Return the database name.	*/

	/* Queries functions */
	void *(*DBexecute_query) (struct db_conn *, char *);	/**< Execute a query.	*/
	int (*DBget_query_status) (struct db_conn *, void *);	/**< Return the status of the query.	*/
	int (*DBget_record_number) (struct db_conn *, void *);	/**< Return the record count.	*/
	int (*DBget_field_number) (struct db_conn *, void *);	/**< Get the field (column) count.	*/
	char *(*DBget_field_name) (struct db_conn *, void *, int field);	/**< Get a field name.	*/
	char *(*DBget_field_value) (struct db_conn *, void *, int record, int field);	/**< Get a field value.	*/
	void (*DBclear_result) (struct db_conn *, void *);	/**< Clear the query result.	*/

	Keyword *keylist;	/**< Keywords for autocompletion.	*/

	/* Tables functions (definitions, ...) */
	int (*DBget_table_def) (struct db_conn *);			/**< Load table definitions.	*/
	void (*DBfree_table_def) (struct db_conn *);		/**< Free table definitions.	*/
	int (*DBget_table_num) (struct db_conn *);			/**< Get the number of tables.	*/
	DBTableDef *(*DBget_table) (struct db_conn *, int);	/**< Get a table definition.	*/

	/* Connection parameters	*/
	char *server;	/**< Server we're connected to.	*/
	char *user;		/**< Username	*/
	char *pass;		/**< password	*/

	void *pvt;	/**< Private data for use of the database driver.	*/
	int pvtlen;	/**< Size of the private data.	*/
};

/** DBConnection is the type used to refer to the db_conn structure.	*/
typedef struct db_conn DBConnection;

/** Data passed from the application to the driver when initiating a
 * connection. Contains login data, and a pointer to a callback function in the
 * application that can display an error message from the driver. Most fields 
 * in this structure are initialised by the database driver in the registration
 * function.	*/
struct _dbconnector {
	char *dbtype;	/**<- Database driver type string.	*/

	/** Callback function for the application to call to open a connection.	*/
	int (*db_connect_callback) (DBConnector *, DBConnection **);
	void (*url_to_form) (Form *, char *);	/**< Load a URL into form fields.	*/
	char *(*form_to_url) (Form *);	/**< Return a URL string from form fields.	*/

	Form *db_frm;	/**< Form listing the fields required to connect.	*/

	/** Function in the application for the driver to call to deliver error
 	 * messages to the user.	*/
	void (*app_error_dialog) (char *, char *);
};

extern DBConnector **connectorTable;

/** Returns an initialised DBConnector for the database driver.	*/
typedef DBConnector *(*db_register_func) (void (*error_dialog) (char *, char *));

/** Connects to the database defined in the current workspace.	*/
DBConnection *db_connect(DBConnector *dbc);

/** Connects to the database specified in th URL.	*/
DBConnection *db_connect_url(char *url);

/** Disconnect from database.	*/
void db_disconnect(DBConnection *conn, void(*display_error) (char *, char *));

/** Allocates a new empty DBConnector.	*/
DBConnector *dbconnector_new(void);
/*@}*/
#endif
