/** \file dialogs.c	Dialog module.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup gui	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2003  Lionel ULMER, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#include <gtk/gtk.h>

#include "appconfig.h"
#include "dialogs.h"
#include "Pics/warning.xpm"
#include "Pics/question.xpm"

/** Call data for dialog.	*/
typedef struct _call_data {
	GtkWidget *dialog;						/**< Dialog widget.	*/
	void (*user_ok_callback) (void *);		/**< Callback for OK button.	*/
	void (*user_cancel_callback) (void *);	/**< Callback for Cancel button.	*/
	void *user_data;						/**< user data.	*/
} CallData;

/** 'OK' callback for connection dialog.	*/
static void dialog_ok_callback(GtkWidget * widget, gpointer data)
{
	CallData *data_rec = (CallData *) data;

	if (data_rec->user_ok_callback != NULL)
		data_rec->user_ok_callback(data_rec->user_data);

	gtk_widget_destroy(GTK_WIDGET(data_rec->dialog));
}

/** Create a standard warning dialog.
 * \param title	The dialog title.
 * \param text	The dialog text.
 * \return	The dialog widget.
 */
GtkWidget *create_warning_dialog(const char *title, const char *text) {
	return create_standard_dialog(title, text, DLGTYPE_WARNING, 
		NULL, "Ok", NULL, NULL, NULL, NULL);
}
	
/** 'Cancel' callback for connection dialog.	*/
static void dialog_cancel_callback(GtkWidget * widget, gpointer data)
{
	CallData *data_rec = (CallData *) data;

	if (data_rec->user_cancel_callback != NULL)
		data_rec->user_cancel_callback(data_rec->user_data);

	gtk_widget_destroy(GTK_WIDGET(data_rec->dialog));
}

/** Creates a default dialog.
 * \param title	The title text (Optional).
 * \param text	The dialog text (Optional).
 * \param predefined_type	One of the DlgType values.
 * \param pixmap	Pixmap to display for DLGTYPE_NONE (NULL if unused).
 * \param ok_text	The text to display on the "OK" button. Optional. If NULL,
 * no OK button will be displayed.
 * \param ok_callback	The callback for the OK button.
 * \param cancel_text	The text for the Cancel button. Optional. If NULL,
 * no Cancel button will be displayed.
 * \param cancel_callback	The callback function for the Cancel button.
 * \param callback_data	The user data to pass to the callback functions.	*/
GtkWidget *create_standard_dialog(const char *title, const char *text,
								  const DlgType predefined_type, char **pixmap,
								  const char *ok_text, void (*ok_callback) (void *),
								  const char *cancel_text,
						 		  void (*cancel_callback) (void *),
								  void *callback_data)
{
	GtkWidget *dialog, *table, *label, *ok_button, *cancel_button;
	CallData *data_send;
	GtkWidget *lbl;

	switch (predefined_type) {
	case DLGTYPE_WARNING:
		pixmap = warning_xpm;
		if (title == NULL)
			title = "Warning";
		break;

	case DLGTYPE_QUESTION:
		pixmap = question_xpm;
		if (title == NULL)
			title = "Question";
		break;

	default:
		break;
	}

	if (title == NULL)
		title = "Message";

	if (text == NULL)
		text = "";

	/* The dialog */
	dialog = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dialog), title);
	gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
	gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
	gtk_grab_add(dialog);
	gtk_widget_realize(dialog);

	/* Now, the top half (pixmap + text) */
	table = gtk_table_new(2, 1, FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(table), 5);
	gtk_table_set_col_spacings(GTK_TABLE(table), 5);

	if (pixmap != NULL) {
		GdkBitmap *mask;
		GdkPixmap *gdkpix;
		GtkStyle *style;
		GtkWidget *gtkpix;

		style = gtk_widget_get_style(dialog);
		gdkpix = gdk_pixmap_create_from_xpm_d(dialog->window,
											  &mask,
											  &style->bg[GTK_STATE_NORMAL],
											  pixmap);
		gtkpix = gtk_pixmap_new(gdkpix, mask);

		gtk_table_attach_defaults(GTK_TABLE(table), gtkpix, 0, 1, 0, 1);
		gtk_widget_show(gtkpix);
	}

	label = gtk_label_new(text);
	gtk_table_attach_defaults(GTK_TABLE(table), label, 1, 2, 0, 1);
	gtk_widget_show(label);

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), table, TRUE,
					   TRUE, 0);
	gtk_widget_show(table);

	/* The structure for the callbacks */
	data_send = (CallData *) g_malloc(sizeof(CallData));
	data_send->dialog = dialog;
	data_send->user_data = callback_data;
	data_send->user_ok_callback = ok_callback;
	data_send->user_cancel_callback = cancel_callback;

	/* The bottom half */
	if (ok_text != NULL) {
		lbl = gtk_label_new(ok_text);
		gtk_misc_set_padding(GTK_MISC(lbl), 4, 0);
		gtk_widget_show(lbl);
		ok_button = gtk_button_new();
		gtk_container_add(GTK_CONTAINER(ok_button), lbl);
		gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
						   GTK_SIGNAL_FUNC(dialog_ok_callback), data_send);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
						   ok_button, FALSE, FALSE, 0);
		GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
		gtk_widget_grab_default(ok_button);
		gtk_widget_show(ok_button);
	}

	if (cancel_text != NULL) {
		lbl = gtk_label_new(cancel_text);
		gtk_misc_set_padding(GTK_MISC(lbl), 4, 0);
		gtk_widget_show(lbl);
		cancel_button = gtk_button_new();
		gtk_container_add(GTK_CONTAINER(cancel_button), lbl);
		gtk_signal_connect(GTK_OBJECT(cancel_button), "clicked",
						   GTK_SIGNAL_FUNC(dialog_cancel_callback),
						   data_send);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
						   cancel_button, FALSE, FALSE, 0);
		gtk_widget_show(cancel_button);
	}

	return dialog;
}


/** Displays an error dialog.	*/
void display_error(char *title, char *message)
{
	GtkWidget *dialog =
		create_standard_dialog(title, message, 1, NULL, "Ok", NULL, NULL,
							   NULL, NULL);
	gtk_widget_show(dialog);
}

/** Tracks whether the open File Dialog is currently displayed.	*/
int openFileDialogDisplayed = 0;

/** Pointer to the open File Dialog widget.	*/
GtkWidget *openFileDialog = NULL;

/** OK callback for the file open dialog. The data pointer is the GtkEntry
 * field to receive the filename.	*/
static void file_open_dialog_ok_callback(GtkWidget * widget, gpointer data)
{
	gchar *filename =
		gtk_file_selection_get_filename(GTK_FILE_SELECTION(openFileDialog));

	cfg_set_current_dir(filename);
	if (data) {
		gtk_entry_set_text(GTK_ENTRY(data), filename);
	}
	openFileDialogDisplayed = 0;
}

/** Display the export dialog.	*/
void file_open_dialog(GtkWidget * txt)
{
	if (openFileDialogDisplayed) {
		/* Should never happen */
	}
	else {
		openFileDialog = gtk_file_selection_new("Select file to open.");
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(openFileDialog),
										cfg_get_current_dir());
		gtk_file_selection_show_fileop_buttons(GTK_FILE_SELECTION
											   (openFileDialog));

		/* Ensure that the dialog box is destroyed when the user clicks a
		 * button. */
		gtk_signal_connect(GTK_OBJECT
						   (GTK_FILE_SELECTION(openFileDialog)->ok_button),
						   "clicked",
						   GTK_SIGNAL_FUNC(file_open_dialog_ok_callback),
						   txt);
		gtk_signal_connect_object(GTK_OBJECT
								  (GTK_FILE_SELECTION(openFileDialog)->
								   ok_button), "clicked",
								  GTK_SIGNAL_FUNC(gtk_widget_destroy),
								  (gpointer) openFileDialog);
		gtk_signal_connect_object(GTK_OBJECT
								  (GTK_FILE_SELECTION(openFileDialog)->
								   cancel_button), "clicked",
								  GTK_SIGNAL_FUNC(gtk_widget_destroy),
								  (gpointer) openFileDialog);

		/* Display that dialog */
		gtk_grab_add(openFileDialog);
		gtk_widget_show(openFileDialog);
		openFileDialogDisplayed = 1;
	}
}

/** Create and show a standard warning dialog.
 * \param title	The dialog title.
 * \param text	The dialog text.
 * \return	The dialog widget.
 */
GtkWidget *show_warning_dialog(char *title, char *text) {
	GtkWidget *dlg = create_standard_dialog(title, text, DLGTYPE_WARNING, NULL, "Ok", NULL, NULL, NULL, NULL);
	if (dlg != NULL)
		gtk_widget_show(dlg);

	return dlg;
}
