/** \file exporthtml.c	Functions to export results to files.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup gui
 *
 * A number of output formats are supported. There are a number of dialogs supported in this module.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2003  Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it andraw/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include "buf.h"
#include "guiapi.h"
#include "queries.h"
#include "status.h"

/** Data for HTML export dialog.	*/
typedef struct _html_export_data {
	GtkWidget *dialog;			/**< Dialog widget.	*/
	char *filename;				/**< Filename with path.	*/
	char *raw_filename;			/**< Raw filename without path.	*/

	GtkWidget *query;			/**< Query widget.	*/
	GtkWidget *recPerPage;		/**< Records per page.	*/
	GtkWidget *titles;			/**< Title string.	*/
	GtkWidget *links;			/**< Links.	*/
} HtmlExportData;

/** Hide the html export dialog.	*/
void export_html_dialog_hide();

/** Show the html export dialog.	*/
void export_html_dialog_show(char *title, char *filename, char *name,
							 GtkWidget * query, char *query_name);

/** Create the html export dialog.	*/
void export_html_dialog_create();

/** Cancel callback for HTML export dialog.	*/
static void export_html_cancel_callback(GtkWidget * widget, gpointer data);

/** OK callback for HTML export dialog.	*/
static void export_html_ok_callback(GtkWidget * widget, gpointer data);

/** HTML export dialog.	*/
static GtkWidget *exportHtmlDialog = NULL;

/** Non-zero if html export dialog is displayed.	*/
static int exportHtmlDialogDisplayed = 0;

/** Data for the html export dialog.	*/
HtmlExportData exportHtmlData = {
	NULL,							   /* GtkWidget *dialog; */
	NULL,							   /* char *filename; */
	NULL,							   /* char *raw_filename; */
	NULL,							   /* GtkWidget *query; */
	NULL,							   /* GtkWidget *rec_per_page; */
	NULL,							   /* GtkWidget *titles; */
	NULL							   /* GtkWidget *links; */
};

/** Create the HTML export dialog. Only needs to be called once.	*/
void export_html_dialog_create()
{
	GtkWidget *table;
	GtkWidget *vbox, *hbox;
	GtkWidget *titles, *links, *label;
	GtkWidget *recPerPage;
	GtkWidget *ok_button, *cancel_button;

	if (exportHtmlDialog)
		return;

	exportHtmlDialog = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(exportHtmlDialog), "Export html...");
	gtk_window_position(GTK_WINDOW(exportHtmlDialog), GTK_WIN_POS_MOUSE);
	gtk_window_set_policy(GTK_WINDOW(exportHtmlDialog), FALSE, FALSE, FALSE);

	vbox = gtk_vbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(exportHtmlDialog)->vbox), vbox,
					   FALSE, FALSE, 0);
	gtk_widget_show(vbox);

	table = gtk_table_new(2, 2, FALSE);

	hbox = gtk_hbox_new(FALSE, 5);
	label = gtk_label_new("Records per page: ");
	gtk_widget_show(label);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, FALSE, 5);

	recPerPage = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(recPerPage), "40");
	gtk_widget_show(recPerPage);

	gtk_box_pack_start(GTK_BOX(hbox), recPerPage, TRUE, FALSE, 5);
	gtk_widget_show(hbox);
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 0, 2, 0, 1);

	links = gtk_check_button_new_with_label("Include links");
	gtk_widget_show(links);
	gtk_table_attach_defaults(GTK_TABLE(table), links, 0, 1, 1, 2);

	titles = gtk_check_button_new_with_label("Include columns titles");
	gtk_widget_show(titles);
	gtk_table_attach_defaults(GTK_TABLE(table), titles, 1, 2, 1, 2);

	gtk_box_pack_start(GTK_BOX(vbox), table, TRUE, FALSE, 5);
	gtk_widget_show(table);

	exportHtmlData.titles = titles;
	exportHtmlData.links = links;
	exportHtmlData.recPerPage = recPerPage;

	ok_button = gtk_button_new_with_label("Ok");
	gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
					   GTK_SIGNAL_FUNC(export_html_ok_callback),
					   &exportHtmlData);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(exportHtmlDialog)->action_area),
					   ok_button, FALSE, FALSE, 0);
	GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(ok_button);
	gtk_widget_show(ok_button);
	cancel_button = gtk_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(cancel_button), "clicked",
					   GTK_SIGNAL_FUNC(export_html_cancel_callback),
					   exportHtmlDialog);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(exportHtmlDialog)->action_area),
					   cancel_button, FALSE, FALSE, 0);
	gtk_widget_show(cancel_button);
	// gtk_widget_show(exportHtmlDialog);
}

/** Hide the HTML export dialog.	*/
void export_html_dialog_hide()
{
	exportHtmlDialogDisplayed = 0;
	gtk_widget_hide(GTK_WIDGET(exportHtmlDialog));
}

/** Show the HTML export dialog.
 * \param title	The title string for the dialog window.
 * \param filename	The selected filename (includes path).
 * \param	rawfilename	The raw filename (no path).
 * \param query	A pointer to the query text widget.
 * \param query_name	The display name of the query.	*/
void export_html_dialog_show(char *title, char *filename, char *rawfilename,
							 GtkWidget * query, char *query_name)
{
	ABuf *buf = NULL;
	void *result;
	DBConnection *conn;

	if (exportHtmlDialogDisplayed)
		return;
	
	if (exportHtmlDialog == NULL) {
		export_html_dialog_create();
		if (exportHtmlDialog == NULL)
			return;
	}

	/* Find the number of rows in the result.	*/
	conn = gui_get_connection();
	result = query_get_result(GTK_WIDGET(query));
	if (conn->DBget_query_status(conn, result) != DB_QUERY_RECORDS_OK)
		return;

	buf = buf_check(buf, 50);
	snprintf(buf->b_dat, buf->b_len, "%d",
			 conn->DBget_record_number(conn, result));
	gtk_entry_set_text(GTK_ENTRY(exportHtmlData.recPerPage), buf->b_dat);

	if (exportHtmlData.filename)
		g_free(exportHtmlData.filename);
	exportHtmlData.filename = g_strdup(filename);
	
	if (exportHtmlData.raw_filename)
		g_free(exportHtmlData.raw_filename);
	exportHtmlData.raw_filename = g_strdup(rawfilename);

	exportHtmlData.query = query;

	gtk_window_set_title(GTK_WINDOW(exportHtmlDialog), title);
	exportHtmlDialogDisplayed = 1;
	gtk_widget_show(GTK_WIDGET(exportHtmlDialog));
}

/** OK callback for HTML export dialog.
 * \param widget	The OK button?
 * \param data	The HtmlExportData structure.	*/
static void export_html_ok_callback(GtkWidget * widget, gpointer data)
{
/** Filename template for the HTML files.	*/
#define HTMLFILE_S	"%s%03d.html"
	HtmlExportData *cd = (HtmlExportData *) data;
	FILE *f = NULL;
	void *result;
	char *basename, *lbase = NULL;
	ABuf *current_file = NULL;
	int currentPage, recPerPage;
	int links, titles;
	int nb_tuples;
	int nb_cols;
	int i, j;
	int col;
	gchar *queryText = NULL;
	DBConnection *conn = gui_get_connection();

	/* Gets all the info we need before exporting */
	result = query_get_result(GTK_WIDGET(cd->query));

	if (conn->DBget_query_status(conn, result) != DB_QUERY_RECORDS_OK)
		return;
	nb_tuples = conn->DBget_record_number(conn, result);
	nb_cols = conn->DBget_field_number(conn, result);

	basename = gtk_entry_get_text(GTK_ENTRY(cd->recPerPage));
	recPerPage = atoi(basename);
	if (recPerPage < 1) {
		g_warning("Invalid records per page\n");
		return;
	}
	basename = NULL;
	links = GTK_TOGGLE_BUTTON(cd->links)->active;
	titles = GTK_TOGGLE_BUTTON(cd->titles)->active;

	if (recPerPage < nb_tuples) {
		basename = strdup(cd->filename);
		for (i = strlen(basename); (i > 0) && (basename[i] != '.'); i--) ;
		basename[i] = '\0';
		for (j = i; (j > 0) && (basename[j] != '/'); j--) ;
		lbase = strdup(basename + j + 1);
	}

	/* Start exporting */
	status_push("Exporting HTML...");
	progress_reset();
	currentPage = 0;

	queryText = query_get_text(GTK_WIDGET(cd->query));

	for (i = 0; i < nb_tuples; i++) {
		if ((i % recPerPage) == 0) {
			if (recPerPage < nb_tuples) {
				if (currentPage != 0) {
					fprintf(f, "</TABLE>\n");
					fclose(f);
				}
				current_file =
					buf_check(current_file,
							  strlen(HTMLFILE_S) + strlen(basename) + 50);
				snprintf(current_file->b_dat, current_file->b_len, HTMLFILE_S,
						 basename, currentPage++);
				f = fopen(current_file->b_dat, "w");
				if (links) {
					if (currentPage != 1)
						fprintf(f, "<A HREF=\"%s%03d.html\">Previous</A> ",
								lbase, currentPage - 2);
					if (nb_tuples > (currentPage + 1) * recPerPage)
						fprintf(f, " <A HREF=\"%s%03d.html\">Next</A>", lbase,
								currentPage);
				}
				fprintf(f, "%s", "<TABLE border>\n");
				if (currentPage == 0)
					fprintf(f, "<CAPTION>%s</CAPTION>\n", queryText);
				else
					fprintf(f, "<CAPTION>%s (continued...)</CAPTION>\n", queryText);

				if (titles) {
					fprintf(f, "%s", " <TR>\n");
					for (col = 0; col < nb_cols; col++) {
						fprintf(f, "  <TH>%s</TH>\n",
								conn->DBget_field_name(conn, result, col));
					}
					fprintf(f, "%s", " </TR>\n");
				}
			}
			else if (i == 0) {
				f = fopen(cd->filename, "w");
				fprintf(f, "%s", "<TABLE border>\n");
				fprintf(f, "<CAPTION>%s</CAPTION>\n", queryText);
				if (titles) {
					fprintf(f, "%s", " <TR>\n");
					for (col = 0; col < nb_cols; col++) {
						fprintf(f, "  <TH>%s</TH>\n",
								conn->DBget_field_name(conn, result, col));
					}
					fprintf(f, "%s", " </TR>\n");
				}
			}
		}

		/* Exports the line */
		fprintf(f, "%s", " <TR>\n");
		for (j = 0; j < nb_cols; j++) {
			fprintf(f, "  <TD> %s </TD>\n",
					conn->DBget_field_value(conn, result, i, j));
		}
		progress_update(i / (float) nb_tuples);
		fprintf(f, "%s", " </TR>\n");
	}

	/* Closes the file */
	fprintf(f, "%s", "</TABLE>\n");
	fclose(f);

	progress_reset();
	status_pop();

	/* Memory cleaning */
	g_free(basename);
	if (queryText != NULL)
		g_free(queryText);

	buf_free(current_file);

	export_html_dialog_hide();
}

/** Cancel callback for HTML export dialog.
 * \param widget	The cancel button.
 * \param data	A pointer to the dialog.	*/
static void export_html_cancel_callback(GtkWidget * widget, gpointer data)
{
	gtk_widget_hide(GTK_WIDGET(data));
}

/** Export results to HTML file.
 * \param filename	The full pathname to output to.
 * \param rawfilename	The raw filename.
 * \param query	The query widget.
 * \param query_name	The query name.	*/
void export_as_html(char *filename, char *rawfilename,
						   GtkWidget * query, char *query_name)
{
/** Template for the HTML export dialog title.	*/
#define EXPORTHTML_S	"Export '%s' as HTML"
	ABuf *buf = NULL;

	export_html_dialog_create();
	buf = buf_new(strlen(EXPORTHTML_S) + strlen(query_name));
	if (!buf) {
		g_error("ExportAsHtml() could not allocate memory for a buffer\n");
		return;
	}

	snprintf(buf->b_dat, buf->b_len, EXPORTHTML_S, query_name);
	export_html_dialog_show(buf->b_dat, filename, rawfilename, query,
							query_name);
	buf_free(buf);
}
