/** \file forms.c	Forms interface.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup forms
 *
 * Each form can have a title and a number
 * of fields. Each field has a name and a value. This is used in GtkSQL as
 * the interface for database drivers to popup login dialogs etc. The driver
 * creates a Form object that defines the fileds for the login dialog. GtkSQL
 * then turns this into a GTK dialog.
 *
 * \par FormField's
 * The FormField type has built-in support for a linked list. Some field functions work on a single field,
 * others work on a field list. All functions take a FormField pointer as the first parameter. The ones that work
 * on a list normally take an integer index as a second parameter. This selects the index number in the list
 * of the field you want to work on. So given a function like fld_xyz(FormField *fld, int n), the call frm_xyz(fld, 5) 
 * would operate on the sixth element of the field list based at 'fld'. The index number is zero-based, so '0' refers
 * to the first field in the list, '1' to the second, etc.
 */
/* $Id: forms.c,v 1.11 2003/09/18 07:50:30 djluff Exp $
 * GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2003 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "buf.h"
#include "forms.h"

/** Longest field name supported.	*/
#define LONGEST_KEY 256

/** Clear the title and all fields from the form.	*/
void frm_clear(Form *frm);

/** Add a field, and set the value, tip and help text.	*/
void frm_define_field(Form *frm, char *nm, char *val);

/** Print a form to stdout for debugging.	*/
void frm_print(Form *frm);

/** Add a field to the end of the list.
 * \param fld	If NULL, allocates a new field and returns it as the list.
 * \param nm	The name for the field.
 * \return A pointer to the list. If 'fld' was NULL, the return value will be the newly allocated Field. If
 * 'fld' was NOT NULL, the return value will be 'fld'.	*/
FormField *fld_add(FormField *fld, char *nm) {
	FormField *fnew = NULL;

	if (! fld) {
		fnew = fld_new(nm);
		return fnew;
	}
	else if (fld->next) {
		/* pass it on	*/
		fld_add((FormField *)(fld->next), nm);
		return fld;
	}
	else {
		fnew = fld_new(nm);
		fld->next = fnew;
		return fld;
	}
	return NULL;
}

/** Count the fields in the list.
 * \param fld	The start of the list.
 * \return The field count.	*/
int fld_count(FormField *fld) {
	FormField *f = fld;
	int cnt = 0;

	while (f) {
		cnt++;
		f = f->next;
	}
	return cnt;
}

/** Frees the field list based at 'fld'.
 * \param fld	The base of the list.
 * \return Always returns NULL. This is so you can do 
 * "fields = fld_free(fields);" to NULL the pointer.	*/
FormField *fld_free(FormField *fld) {
	if (! fld)
		return NULL;

	if (fld->next)
		fld_free(fld->next);

	fld_set_name(fld, NULL);
	fld_set_value(fld, NULL);
	free(fld);
	return NULL;
}

/** Retrieve the help text for the field.
 * \param	fld	The start field.
 * \param	i	The index of the field to find.
 * \return	A pointer to the selected field.	*/
FormField *fld_get(FormField *fld, int i) {
	if (! fld)
		return NULL;

	if (! i)
		return fld;
	else if (fld->next)
		return fld_get(fld->next, i-1);
	else
		return NULL;
}

/** Retrieve the help text for the field.
 * \param	fld	The field.
 * \return	The help string, or NULL if none specified.	*/
char *fld_get_help(FormField *fld) {
	if (! fld)
		return NULL;

	return fld->help;
}

/** Retrieve the tooltip text for the field.
 * \param	fld	The field.
 * \return	The relevance option.	*/
FieldRelevance fld_get_relevance(FormField *fld) {
	if (! fld)
		return FR_OPTIONAL;

	return fld->rel;
}

/** Retrieve the tooltip text for the field.
 * \param	fld	The field.
 * \return	The text string, or NULL if none specified.	*/
char *fld_get_tip(FormField *fld) {
	if (! fld)
		return NULL;

	return fld->tip;
}

/** Returns the edit mode of the selected field in the list based at 'fld'.
 * \param fld	The field.
 * \return The field type.
 * \see FieldType
 * \see FT_NONE
 * \see FT_TEXT
 * \see FT_PWD
 * \see FT_FILENAME	*/
FieldType fld_get_type(FormField *fld) {
	if (! fld)	/* nothing there	*/
		return FT_NONE;

	return fld->type;
}

/** Get the name of the field.
 * \param fld
 * \return The name pointer. Dont free this pointer.	*/
char *fld_get_name(FormField *fld) {
	if (! fld)	/* nothing there	*/
		return NULL;

	return fld->name;
}

/** Get the value of the selected field in the list based at 'fld'.
 * \param fld	The base of the field list.
 * \return The field value string.	*/
char *fld_get_value(FormField *fld) {
	if (! fld)
		return NULL;

	return fld->value;
}

/** Returns the value of field with the name 'nm'.	*/
char *fld_get_value_by_name(FormField *fld, char *nm) {
	if (! fld)
		return NULL;

	if ((fld->name) && (!strcmp(nm, fld->name)))
		return fld->value;

	if (fld->next)
		return fld_get_value_by_name(fld->next, nm);

	return NULL;
}

/** Creates and returns a new field.	*/
FormField *fld_new(char *nm) {
	FormField *fnew = (FormField *)calloc(1, sizeof(FormField));
	if (fnew) {
		fnew->type = FT_TEXT;
		fnew->rel = FR_OPTIONAL;
		if (nm)
			fnew->name = strdup(nm);
	}
	return fnew;
}

/** Returns the next field in the list.	*/
FormField *fld_next(FormField *fld) {
	if (! fld)
		return NULL;
	else
		return fld->next;
}

/** Sets the edit mode for the specified field.
 * \param fld	The field to modify.
 * \parm typ The new type.	*/
void fld_set_type(FormField *fld, FieldType typ) {
	if (! fld)
		return;

	fld->type = typ;
}

/** Set the help text for the field.
 * \param	fld	The field to set the help on.
 * \param	hlp	The help text.	*/
void fld_set_help(FormField *fld, char *hlp) {
	if (! fld)
		return;

	if (fld->help)
		free(fld->help);

	if (hlp)
		fld->help = strdup(hlp);
}

/** Set the relevance of the field.
 * \param	fld	The field to set the help on.
 * \param	r	The relevance to set.	*/
void fld_set_relevance(FormField *fld, FieldRelevance r) {
	if (! fld)
		return;

	fld->rel = r;
}

/** Set the 'tooltip' text for the field.
 * \param	fld	The field to set the help on.
 * \param	txt	The tooltip text.	*/
void fld_set_tip(FormField *fld, char *txt) {
	if (! fld)
		return;

	if (fld->tip) {
		free(fld->tip);
		fld->tip = NULL;
	}

	if (txt)
		fld->tip = strdup(txt);
}

/** Sets the name of the specified field.	*/
FormField *fld_set_name(FormField *fld, char *nm) {
	if (! fld)
		return NULL;

	if (fld->name)
		free(fld->name);

	if (nm)
		fld->name = strdup(nm);
	else
		fld->name = NULL;

	return fld;
}

/** Sets the value of the specified field.
 * \param fld The field to set.
 * \param val The new value (may be NULL).	*/
void fld_set_value(FormField *fld, char *val) {
	if (! fld)
		return;

	if (fld->value)
		free(fld->value);

	if (val)
		fld->value = strdup(val);
	else
		fld->value = NULL;
}

/** Sets a field value.	*/
char *fld_set_value_by_name(FormField *fld, char *nm, char *val) {
	if (! fld)
		return NULL;

	if ((fld->name) && (!strcmp(nm, fld->name)))
		fld_set_value(fld, val);
	else if (fld->next)
		return fld_set_value_by_name(fld->next, nm, val);

	return NULL;
}

/** Adds a form field. Returns a pointer to the added field.	*/
FormField *frm_add_field(Form *frm, char *nm, char *val) {
	FormField *f;
	if (! frm)
		return NULL;

	frm->fld = fld_add(frm->fld, nm);
	f = frm_last_field(frm);
	if (f)
		fld_set_value(f, val);
	return f;
}

/** Clone an existing form. All fields and field values are copied to the new
 * form.
 * \param frm The form to clone.
 * \return The newly allocated form. If any field cannot be copied, NULL
 * is returned.	*/
Form *frm_clone(Form *frm) {
	Form *f;

	if (! frm)
		return NULL;

	f = frm_new(frm);
	return f;
}


/** Copies one form to another.
 * \param dst The destination form.
 * \param src The source form.	*/
void frm_copy(Form *dst, Form *src) {
	FormField *fld;
	int fields;
	int i;
	char *name;
	char *val;

	if ((! dst) || (! src))
		return;

	frm_clear(dst);
	frm_set_title(dst, frm_get_title(src));

	fields = frm_get_field_count(src);
	for (i=0; i<fields; i++) {
		fld = frm_get_field(src, i);
		if (fld) {
			name = fld_get_name(fld);
			if (name) {
				val = fld_get_value_by_name(fld, name);
				frm_add_field(dst, name, val);
			}
		}
	}
}

/** Creates a pane from a form.
 * \param	frm	The form data to use.
 * \return A Buf containing an error message, or NULL if no error.
 * If a Buf is returned, the caller should free it after use.	*/
ABuf *frm_check_fields(Form * frm)
{
	FormField *fld;

	int i;
	int fieldcount;

	char *value;
	ABuf *errmsg = NULL;
	FieldRelevance rel;

	/* Creates the notebook pane and label */
	/* The connection pane */
	fieldcount = frm_get_field_count(frm);
	for (i = 0; i < fieldcount; i++) {
		fld = frm_get_field(frm, i);
		rel = fld_get_relevance(fld);
		if (rel == FR_REQUIRED) {
			value = fld_get_value(fld);
			if ((value == NULL) || (strlen(value) == 0)) {
				/* No good.	*/
				errmsg = buf_strcat(errmsg, fld_get_name(fld));
				errmsg = buf_strcat(errmsg, ", ");
			}
		}
	}
	if (errmsg != NULL) {
		/* Remove the trainling ", "	*/
		errmsg = buf_str_shrinkbyone(errmsg);
		errmsg = buf_str_shrinkbyone(errmsg);
	}
	return errmsg;
}

/** Deletes all the fields in a form. Internal function.
 * \param frm The form to operate on.	*/
void frm_clear(Form *frm) {
	frm_set_title(frm, NULL);
	fld_free(frm->fld);
	return;
}

/** Clears the values of all fields, but doesn't free them.
 * \param frm The form to operate on.	*/
void frm_clear_fields(Form *frm) {
	FormField *fld;

	if (frm == NULL)
		return;

	fld = frm->fld;
	while (fld != NULL) {
		fld_set_value(fld, NULL);
		fld = fld->next;
	}
}

/** Frees the memory for a form.	*/
Form *frm_free(Form *f) {
	frm_set_title(f, NULL);
	fld_free(f->fld);
	free(f);
	return NULL;
}

/** Returns the edit mode of the field 'n'.	*/
FormField *frm_get_field(Form *frm, int n) {
	if ((! frm) || (! frm->fld))
		return NULL;

	return fld_get(frm->fld, n);
}

/** Returns the edit mode of the field 'n'.	*/
FieldType frm_get_field_type(Form *frm, int n) {
	FormField *fld;

	if ((! frm) || (! frm->fld))
		return FT_NONE;

	fld = frm_get_field(frm, n);
	if (fld != NULL)
		return fld_get_type(fld);

	return FT_NONE;
}

/** Returns the count of active fields.	*/
int frm_get_field_count(Form *frm) {
	if ((! frm) || (! frm->fld))
		return 0;
	
	return fld_count(frm->fld);
}

/** Returns the name of the selected field.	*/
char *frm_get_field_name(Form *frm, int i) {
	FormField *fld;

	if ((! frm) || (! frm->fld))
		return NULL;

	fld = frm_get_field(frm, i);
	return fld_get_name(fld);
}

/** Returns a pointer to the last added field.	*/
FormField *fld_get_last(FormField *fld) {
	FormField *f = fld;

	if (! f)
		return NULL;

	while (f->next)
		f = f->next;

	return f;
}

/** Returns the form title.	*/
char *frm_get_title(Form *frm) {
	if (! frm)
		return NULL;

	return frm->title;
}

/** Return the value of the selected field.
 * \param frm	The form to query.
 * \param i	The zero-based index.
 * \return	The value, or NULL.	*/
char *frm_get_value(Form *frm, int i) {
	FormField *fld;

	if (! frm)
		return NULL;

	if (! frm->fld)
		return NULL;

	fld = frm_get_field(frm, i);
	if (fld == NULL)
		return NULL;
	else
		return fld_get_value(fld);
}

/** Returns the value of field with the name 'nm'.	*/
char *frm_get_value_by_name(Form *frm, char *nm) {
	return fld_get_value_by_name(frm->fld, nm);
}

/** Returns a pointer to the last added field.	*/
FormField *frm_last_field(Form *frm) {
	if ((! frm) || (! frm->fld))
		return NULL;

	return fld_get_last(frm->fld);
}

/** Load a Form file from disk.
 * \param frm	The form to load to.
 * \param fn	Filename to load. If NULL, returns without doing anything, otherwise
 * clears the current form fields and attempts to load the file. If there's an
 * error, sets the "frm_error" field to an error string.
 * \return Zero on success, non-zero on error.	*/
int frm_load_file(Form *frm, char *fn)
{
	FILE *ff;
	char buf[1024];
	char key[LONGEST_KEY + 1];
	char *val;

	if ((frm == NULL) || (fn == NULL))
		return 1;

	frm_clear(frm);

	ff = fopen(fn, "r");
	if (!ff) {
		perror("opening form file for reading");
		frm_set_value_by_name(frm, FRM_ERROR, "Could not open form file");
		return 1;
	}

	while (fgets(buf, sizeof(buf), ff)) {
		char *colon = strchr(buf, ':');
		if (colon) {
			*colon = '\0';
			strcpy(key, buf);
			val = colon;
			val++;

			/* Skip leading white space */
			while ((*val) && (isspace(*val)))
				val++;

			if (strlen(val) > 0) {
				frm_define_field(frm, key, val);
			}
		}
	}
	frm_print(frm);
	return 0;
}

/** Compares the two forms.
 * \param frm1	The first form
 * \param frm2	The second form.
 * \return Zero for a non-match, non-zero for a match.	*/
int frm_match(Form *frm1, Form *frm2) {
	char *name;
	char *val1;
	char *val2;
	int i;
	int fields1 = frm_get_field_count(frm1);
	int fields2 = frm_get_field_count(frm2);

	if (fields1 != fields2)
		return 0;

	for (i=0; i<fields1; i++) {
		name = frm_get_field_name(frm1, i);
		val1 = frm_get_value(frm1, i);
		val2 = frm_get_value_by_name(frm2, name);
		if (val1 == NULL) {
			if (val2 != NULL)
				return 0;
		}
		else {
			if (val2 == NULL)
				return 0;
			if (! strcmp(val1, val2))
				return 0;
		}
	}
	return 1;
}

/** Creates and returns a new form.	*/
Form *frm_new(Form *frm) {
	Form *f = calloc(1, sizeof(Form));
	if (frm)
		frm_copy(f, frm);

	return f;
}

/** Print the form values for debugging.
 * \param frm	The form to print.	*/
void frm_print(Form *frm) {
	FormField *fld;
	if (frm == NULL)
		return;

	printf("Title: %s\n", frm_get_title(frm));
	printf("=Fields=\n");
	fld = frm->fld;
	while (fld != NULL) {
		printf(" %s: %s|%s|%s\n", fld_get_name(fld), fld_get_value(fld), fld_get_tip(fld), fld_get_help(fld));
	}
}

/** Save the Form to a file.
 * \param	frm	The form to save.
 * \param	fn	The filename to save the file to.
 * \return	An integer error code. -1 = error opening file. 0 = success.	*/
int frm_save_file(Form *frm, char *fn)
{
	FILE *ff;
	FormField *fld;
	char *s;
	ABuf *buf = buf_new(256);

	if ((frm == NULL) || (fn == NULL))
		return 1;

	ff = fopen(fn, "w");
	if (!ff) {
		perror("opening workspace file for writing");
		frm_set_value_by_name(frm, FRM_ERROR, "Can't open form file for writing.");
		return -1;
	}

	s = frm_get_title(frm);
	if (s != NULL)
		fprintf(ff, "Title: %s\n", s);

	fld = frm->fld;
	while (fld != NULL) {
		s = fld_get_name(fld);
		if (s) {
			buf = buf_strcpy(buf, s);
			buf = buf_strcat(buf, ": ");
			s = fld_get_value(fld);
			if (s == NULL)
				s = "";
			buf = buf_strcat(buf, s);
			s = fld_get_tip(fld);
			if ((s != NULL) && (strlen(s) > 0)) {
				buf = buf_strcat(buf, "|");
				buf = buf_strcat(buf, s);
				s = fld_get_help(fld);
				if ((s != NULL) && (strlen(s) > 0)) {
					buf = buf_strcat(buf, "|");
					buf = buf_strcat(buf, s);
				}
			}
			fprintf(ff, "%s\n", buf->b_dat);
		}
		fld = fld_next(fld);
	}
	buf_free(buf);
	return 0;
}

/** Sets the edit mode for the specified field.	*/
void frm_set_field_type(Form *frm, int id, int typ) {
	FormField *fld;

	if ((! frm) || (! frm->fld))
		return;

	fld = frm_get_field(frm, id);
	if (fld != NULL)
		fld_set_type(fld, typ);
}

/** Sets the title for a form.	*/
Form *frm_set_title(Form *f, char *ttl) {
	if (! f)
		return NULL;

	if (f->title) {
		free(f->title);
		f->title = NULL;
	}

	f->title = strdup(ttl);
	return f;	
}

/** Sets a field value.	*/
void frm_set_value_by_name(Form *frm, char *nm, char *val) {
	fld_set_value_by_name(frm->fld, nm, val);
}

/** Sets a field value from a form file line.
 * \param frm	The form to operate on.
 * \param nm	The field name.
 * \param val	The field value, tip text(optional) and help text (optional).
 * These are in the format:
 * value[|hint[|help]].	*/
void frm_define_field(Form *frm, char *nm, char *val) {
	FormField *fld;
	char *value = NULL;
	char *tip = NULL;
	char *help = NULL;
	char *pipe = NULL;
	char *start = val;

	/* Find value.	*/
	pipe = strchr(start, '|');
	value = val;
	if (pipe == NULL) {
		/* Whole string must be value.	*/
		goto done;
	}

	/* terminate value string.	*/
	*pipe = '\0';

	/* Look for tip.	*/
	start = pipe;
	start++;

	if (strlen(start) == 0)
		/* No tip text.	*/
		goto done;

	tip = start;

	/* Look for help	*/
	pipe = strchr(start, '|');
	if (pipe == NULL) {
		/* Rest of string Whole string must be tip.	*/
		goto done;	
	}

	/* terminate tip	*/
	*pipe = '\0';
	
	start = pipe;
	start++;

	if (strlen(start) == 0)
		goto done;

	help = start;

done:
	fld = frm_add_field(frm, nm, val);
	if (tip != NULL)
		fld_set_tip(fld, tip);

	if (help != NULL)
		fld_set_help(fld, help);
}
