/** \addtogroup forms
 *@{
 */
/** \file forms.h	Form interface.
 * \author Lionel ULMER
 * \author Darryl Luff	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2003 Darryl Luff
 * MySQL "driver" (C) 1998 Des Herriott (des@ops.netcom.net.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef __FORMS_H
#define __FORMS_H

#include "buf.h"

/** Form Field type.	*/
typedef enum {
	FT_NONE,	/**<- Invalid field type.	*/
	FT_TEXT,	/**<- Field is a normal text field (default).	*/
	FT_PWD,		/**<- Field is a password field. Typed characters should be hidden in the user interface.	*/
	FT_FILENAME	/**<- Field is for a filename.	*/
} FieldType;

/** Field Relevance. Marks the field as optional or compulsory.	*/
typedef enum {
	FR_OPTIONAL,		/**<- Field is optional (default).	*/
	FR_REQUIRED		/**<- Field must be completed.	*/
} FieldRelevance;

/** The field name that holds form error messages.	*/
#define FRM_ERROR "frm_error"

/** Structure describing a single form field.	*/
typedef struct {
	char *name;		/**< Field name. Used for display.	*/
	FieldType type;	/**< Field type.	*/
	FieldRelevance rel;	/**< Field relevance.	*/
	char *value;		/**< Field value.	*/
	char *tip;		/**< Short tip for 'tooltips' type applications.	*/ 
	char *help;		/**< Longer help text.	*/
	void *next;		/**<- Link to next field in list.	*/
} FormField;

/** Structure describing a single form.	*/
typedef struct {
	/** Form title.	*/
	char *title;
	/** List of fields for this form.	*/
	FormField *fld;
} Form;

/** \brief Add a new field with the given name.	*/
FormField *fld_add(FormField *fld, char *nm);

/** \brief Get the number of form fields.	*/
int fld_count(FormField *fld);

/** \brief Free the specified field and return NULL.	*/
FormField *fld_free(FormField *fld);

/** \brief Get the 'i'th field in the list rooted at 'fld'.	*/
FormField *fld_get(FormField *fld, int i);

/** \brief Retrieve the help text for the subject field.	*/
char *fld_get_help(FormField *fld);

/** \brief Find the last field in the list rooted at 'fld'	*/
FormField *fld_get_last(FormField *fld);

/** \brief Get the name of the specified field.	*/
char *fld_get_name(FormField *fld);

/** Get the relevance of this field.	*/
FieldRelevance fld_get_relevance(FormField *fld);

/** \brief Get the tip text for the specified field.	*/
char *fld_get_tip(FormField *fld);

/** \brief Get the field type.	*/
FieldType fld_get_type(FormField *fld);

/** \brief Get the value text from the field.	*/
char *fld_get_value(FormField *fld);

/** \brief Search the field list rooted at 'fld', and return the value of the field
 * with the name 'nm'.	*/
char *fld_get_value_by_name(FormField *fld, char *nm);

/** \brief Allocate a new field named 'nm'.	*/
FormField *fld_new(char *nm);

/** \brief Return the next field in the list after 'fld'.	*/
FormField *fld_next(FormField *fld);

/** \brief Set the help text for the field.	*/
void fld_set_help(FormField *fld, char *hlp);

/** \brief Set the name of the field.	*/
FormField *fld_set_name(FormField *fld, char *nm);

/** Set the relevance of the field.	*/
void fld_set_relevance(FormField *fld, FieldRelevance r);

/** \brief Set the tip text for the field.	*/
void fld_set_tip(FormField *fld, char *txt);

/** \brief Set the type of the field.	*/
void fld_set_type(FormField *fld, FieldType typ);

/** \brief Set the value of the field.	*/
void fld_set_value(FormField *fld, char *val);

/** \brief Search the field list rooted at 'fld', and set the value of the field(s)
 * with the name 'nm' to 'val'.	*/
char *fld_set_value_by_name(FormField *fld, char *nm, char *val);

/** Creates a pane from a form.	*/
ABuf *frm_check_fields(Form * frm);

/** Clears the values of all fields, but doesn't free them.
 * \param frm The form to operate on.	*/
void frm_clear_fields(Form *frm);

/** \brief Free the specified form.	*/
Form *frm_free(Form *f);

/** \brief Add a new field to the form.	*/
FormField *frm_add_field(Form *frm, char *nm, char *val);

/** \brief Get the 'n'th field from the form.	*/
FormField *frm_get_field(Form *frm, int n);

/** \brief Get the type of the 'n'th field from the form.	*/
FieldType frm_get_field_type(Form *frm, int n);

/** \brief Get the number of fields in the form.	*/
int frm_get_field_count(Form *frm);

/** \brief Get the name of the 'i'th field on the form.	*/
char *frm_get_field_name(Form *frm, int i);

/** \brief Get the form title.	*/
char *frm_get_title(Form *frm);

/** \brief Get the value of the 'i'th field on the form.	*/
char *frm_get_value(Form *frm, int i);

/** \brief Get the value of the form field with the name 'nm'.	*/
char *frm_get_value_by_name(Form *frm, char *nm);

/** \brief Get the last field on the form.	*/
FormField *frm_last_field(Form *frm);

/** \brief Compare two forms to see if they are identical.	*/
int frm_match(Form *frm1, Form *frm2);

/** \brief Allocate a new form.	*/
Form *frm_new(Form *frm);

/** Save the Form to a file.	*/
int frm_save_file(Form *frm, char *fn);

/** \brief Set the type of the 'i'th field on the form.	*/
void frm_set_field_type(Form *frm, int i, int typ);

/** \brief Set the title of the form.	*/
Form *frm_set_title(Form *f, char *ttl);

/** \brief Set the value of the form field(s) named 'nm'.	*/
void frm_set_value_by_name(Form *frm, char *nm, char *val);

/*@}*/
#endif
