/** \file gui.c	Implements the GUI API.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup gui	*/
/* $Id: gui.c,v 1.9 2003/09/14 19:25:12 djluff Exp $
 * GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2003 Lionel Ulmer, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#include <string.h>
#include "buf.h"
#include "callbacks.h"
#include "dialogs.h"
#include "export.h"
#include "guiapi.h"
#include "luaif.h"
#include "mytoolbar.h"
#include "queries.h"
#include "status.h"
#include "support.h"
#include "tables.h"
#include "workspace.h"

/* Icons for the toolbar */
#include "../Pics/connect.xpm"
#include "../Pics/disconnect.xpm"
#include "../Pics/send.xpm"
#include "../Pics/new.xpm"
#include "../Pics/save.xpm"
#include "../Pics/open.xpm"
#include "../Pics/delete.xpm"
#include "../Pics/rename.xpm"
#include "../Pics/export.xpm"
#include "../Pics/refresh.xpm"

#include "../Pics/connect-grey.xpm"
#include "../Pics/disconnect-grey.xpm"
#include "../Pics/send-grey.xpm"
#include "../Pics/new-grey.xpm"
#include "../Pics/save-grey.xpm"
#include "../Pics/open-grey.xpm"
#include "../Pics/delete-grey.xpm"
#include "../Pics/rename-grey.xpm"
#include "../Pics/export-grey.xpm"
#include "../Pics/refresh-grey.xpm"


/** Currently active connection.	*/
DBConnection *guiConnection = NULL;

/** Name of the rc file.	*/
static ABuf *rcname = NULL;

/* Widgets	*/
/** Database list tree.	*/
GtkWidget *tree = NULL;

/** Used to activate/deactivate controls as the application state changes.	*/
typedef struct _activator {
	GtkWidget **widget;		/**< The widget to activate.	*/
	int (*activator) ();	/**< Test function.	If this function returns non-zero, the widget is activated.	*/
} Activator;

/* Buttons	*/
GtkWidget *toolbarAddButton;			/**< 'Add' query button.	*/
GtkWidget *toolbarConnectButton;		/**< 'Connect' button.	*/
GtkWidget *toolbarCloseButton;			/**< 'Close' query Button.	*/
GtkWidget *toolbarDisconnectButton;		/**< 'Disconnect' Button.	*/
GtkWidget *toolbarExportButton;			/**< 'Export' Button.	*/
GtkWidget *toolbarLoadButton;			/**< 'Load' query Button.	*/
GtkWidget *toolbarRefreshButton;		/**< 'Refresh' Button.	*/
GtkWidget *toolbarSaveButton;			/**< 'Save' query Button.	*/
GtkWidget *toolbarSaveAsButton;			/**< 'Save As' query Button.	*/
GtkWidget *toolbarSendButton;			/**< 'Send' query Button.	*/

/** Toolbar definition.	*/
ToolBar toolbardef[] = {
	/* Connection operations. */
	{"Connect", connect_xpm, connect_grey_xpm,
	 "Connect to a new database", on_databaseConnect_activate, NULL,
	 &toolbarConnectButton},

	{"Disconnect", disconnect_xpm, disconnect_grey_xpm,
	 "Disconnect from the current database", on_databaseDisconnect_activate,
	 NULL,
	 &toolbarDisconnectButton},

	{"", NULL, NULL, NULL, NULL, NULL},

	/* Query operations. */
	{"Send", send_xpm, send_grey_xpm,
	 "[Alt-S] Send the query to the database", on_querySend_activate, NULL,
	 &toolbarSendButton},

	{"New", new_xpm, new_grey_xpm,
	 "[Alt-N] Add a new query", on_queryNew_activate, NULL, &toolbarAddButton},

	{"Load", open_xpm, open_grey_xpm,
	 "Load a query from disk", on_queryLoad_activate, NULL,
	 &toolbarLoadButton},

	{"Save", save_xpm, save_grey_xpm,
	 "Save the current query", on_querySave_activate, NULL,
	 &toolbarSaveButton},

	{"Save As", rename_xpm, rename_grey_xpm,
	 "Save query under a new name", on_querySaveAs_activate, NULL,
	 &toolbarSaveAsButton},

	{"Close", delete_xpm, delete_grey_xpm,
	 "[Alt-D] Close current query", on_queryClose_activate, NULL,
	 &toolbarCloseButton},

	{"", NULL, NULL, NULL, NULL, NULL},

	/* Result operations. */
	{"Export", export_xpm, export_grey_xpm,
	 "[Alt-E] Export the current query results", do_export_dialog, NULL,
	 &toolbarExportButton},

	{"", NULL, NULL, NULL, NULL, NULL},

	/* Table operations. */
	{"Refresh", refresh_xpm, refresh_grey_xpm,
	 "Refresh the table display", on_tableRefresh_activate, NULL,
	 &toolbarRefreshButton},

	{NULL, NULL, NULL, NULL, NULL, NULL}
};

static int gui_connection_bad(void);
static int query_exists(void);
static int result_export_ok(void);

/** The activation list of widgets to be activated/deactivated.	*/
Activator sensWidget[] = {
	{&mainmenu_databaseConnect, gui_connection_bad},
	{&mainmenu_databaseDisconnect, gui_connection_ok},
	{&mainmenu_fileQueryImport, query_exists},
	{&mainmenu_fileQueryLoad, gui_connection_ok},
	{&mainmenu_fileQuerySave, query_exists},
	{&mainmenu_fileQuerySaveAs, query_exists},
	{&mainmenu_fileResultExport, result_export_ok},
	{&mainmenu_fileScriptLoad, lua_supported},
	{&mainmenu_fileScriptSave, gui_connection_ok},
	{&mainmenu_queryClear, query_exists},
	{&mainmenu_queryClose, query_exists},
	{&mainmenu_queryNew, gui_connection_ok},
	{&mainmenu_querySend, query_exists},
	{&mainmenu_tableRefresh, gui_connection_ok},
	{&queryPopupClearItem, query_exists},
	{&queryPopupCloseItem, query_exists},
	{&queryPopupImportItem, query_exists},
	{&queryPopupNewItem, gui_connection_ok},
	{&queryPopupSaveItem, query_exists},
	{&queryPopupSaveAsItem, query_exists},
	{&queryPopupSendItem, query_exists},
	{&toolbarAddButton, gui_connection_ok},
	{&toolbarConnectButton, gui_connection_bad},
	{&toolbarCloseButton, query_exists},
	{&toolbarDisconnectButton, gui_connection_ok},
	{&toolbarExportButton, result_export_ok},
	{&toolbarLoadButton, gui_connection_ok},
	{&toolbarRefreshButton, gui_connection_ok},
	{&toolbarSaveButton, query_exists},
	{&toolbarSaveAsButton, query_exists},
	{&toolbarSendButton, query_exists},
	{NULL, NULL}
};


/** Create toolbar.	*/
static GtkWidget *create_toolbar(GtkWidget * mainvbox)
{
	GtkWidget *toolbar;
	int tbc = GTK_TOOLBAR_BOTH;

	toolbar = create_my_toolbar(mainvbox, toolbardef);
	gtk_widget_show(toolbar);

	set_my_toolbar_config(tbc);
	update_button_sensitivity();
	return toolbar;
}

/** Callback for the 'Disconnect' dialog.
 * Resets all the display items.	*/
void disconnect_callback(void *data)
{
	gui_disconnect();
	gtk_window_set_title(GTK_WINDOW(mainWindow), "GtkSQL");
	update_button_sensitivity();
}

/** Checks whether currently connected. If so, it displays 
 * a user a message asking them to disconnect first.
 * \return If the user is connected, return 0. Otherwise,
 * return 1.	*/
int disconnect_first(void)
{
/** Disconnection warning text.	*/
#define DISCONNECTFIRST_S	" You must disconnect first \n from database %s."
	ABuf *buf;
	DBConnection *conn = gui_get_connection();
	if (conn == NULL) {
		return 1;
	}
	else {
		buf =
			buf_new(strlen(DISCONNECTFIRST_S) +
					strlen(conn->DBget_db_name(conn)));
		if (!buf)
			g_warning
				("disconnectFirst() could not allocate memory for a buffer\n");
		else {
			snprintf(buf->b_dat, buf->b_len, DISCONNECTFIRST_S,
					 conn->DBget_db_name(conn));
			display_error("Disconnect", buf->b_dat);
			buf_free(buf);
		}
	}
	return 0;
}

/** Get the name of the rc file to use.
 * \return The filename.	*/
char *get_rcname(void)
{
	rcname = buf_strcat(rcname, getenv("HOME"));
	rcname = buf_strcat(rcname, "/.gtksqlrc");
	return rcname->b_dat;
}

/** Load the current workspace.	*/
void gui_get_workspace()
{
	/* Load the workspace from the file. */
	ws_load_file(NULL);
}

/** Test if connection is bad.
 * \return	non-zero if the connection is invalid.	*/
static int gui_connection_bad(void)
{
	DBConnection *conn = gui_get_connection();

	if (conn == NULL)
		return 1;
	else
		return 0;
}

/** Initialise top hpane. This contains the query pane on the left, and the 
 * table pane on the right.
 * \return	The hpane ready for packing.	*/
GtkWidget *init_top_hpane()
{
	GtkWidget *hpane = lookup_widget(mainWindow, "hpane");

	gtk_paned_add1(GTK_PANED(hpane), query_pane_init());
	gtk_paned_add2(GTK_PANED(hpane), init_tables_pane());

	gtk_widget_show(hpane);
	return hpane;
}

/** Initialise the main VPane. This contains the HPane for the query and
 * table panes at the top, and the result pane at the bottom.
 * \return	The mainvpane ready for packing.	*/
GtkWidget *init_mainvpane()
{
	GtkWidget *mainvpane = lookup_widget(mainWindow, "mainvpane");

	gtk_paned_add1(GTK_PANED(mainvpane), init_top_hpane());
	gtk_paned_add2(GTK_PANED(mainvpane), result_pane_init());

	gtk_widget_show(mainvpane);
	return mainvpane;
}

/** Initialise the status bar.
 * \return The intiialised status bar.	*/
static GtkWidget *init_statusbar(void)
{
	GtkWidget *statusbar = status_bar_new();
	status_push("Welcome to " APPNAME " version " VERSION);
	progress_reset();
	gtk_widget_show(statusbar);
	return statusbar;
}

/** Add the widgets. Must be called after the vbox has been added to the main window.	*/
GtkWidget *populate_mainvbox(GtkWidget * mainvbox)
{
	GtkWidget *tb;
	if (mainvbox == NULL)
		return NULL;

	tb = create_toolbar(mainvbox);
	gtk_box_pack_start(GTK_BOX(mainvbox), tb, FALSE, FALSE, 0);
	gtk_box_reorder_child(GTK_BOX(mainvbox), tb, 1);

	/* Build main hpane including server tree. */
	init_mainvpane();

	gtk_box_pack_end(GTK_BOX(mainvbox), init_statusbar(), FALSE, FALSE, 0);
	return mainvbox;
}

/** Initialise the main window. The gui structure is:
 * mainWindow		(Glade)
 * .mainvbox		(Glade)
 * ..menubar		(Glade)
 * ..toolbar		(Manual)
 * ..mainvpane	(Glade)
 * ...hpane		(Glade)
 * ....query		(Manual)
 * ....tables		(Manual)
 * ...result		(Manual)
 * ..statusbar		(Glade)	*/
void populate_mainWindow()
{
	/* Top level layout. Have to split mainvbox creation and population
	 * because pixmaps can't be created properly until their container is
	 * added to the window. */
	GtkWidget *mainvbox = lookup_widget(mainWindow, "mainvbox");
	if (!mainvbox) {
		g_warning("mainvbox = NULL\n");
	}
	populate_mainvbox(mainvbox);
}

/** Check whether any queries exist.
 * \return	1 if a query exists, 0 otherwise.	*/
static int query_exists(void)
{
	DBConnection *conn = gui_get_connection();
	if (conn == NULL)
		return 0;

	if (query_count() > 0)
		return 1;

	return 0;
}

/** Check whether it's OK to do a result export.
 * \return '1' if OK to procedd.	*/
static int result_export_ok(void)
{
	DBConnection *conn = gui_get_connection();
	if (conn == NULL)
		return 0;

	if (query_count() > 0)
		return !(export_in_progress());

	return 0;
}

/** Make the controls 'sensitive' state match the application state. */
void update_button_sensitivity(void)
{
	int i = 0;
	while (sensWidget[i].widget != NULL) {
		int act = (sensWidget[i].activator) ();
		GtkWidget *w = *(sensWidget[i].widget);
		if (w != NULL)
			gtk_widget_set_sensitive(w, act);
		i++;
	}
}

#ifdef USE_EXPERIMENTAL
/** Initialise the tree.
 * \return The tree widget.	*/
GtkWidget *init_tree()
{
	GtkTreeItem *item;
	GtkWidget *subtree;
	GtkWidget *dbitem;
	GtkWidget *label;
	static gchar *itemtext1[] = { "server1" };
	static gchar *itemtext2[] = { "andrew" };
	static gchar *dbname1[] = { "logs", "racedb" };

	tree = gtk_ctree_new(1, 0);
	gtk_widget_show(GTK_WIDGET(tree));

	subtree = gtk_ctree_new(1, 0);
	gtk_widget_show(GTK_WIDGET(subtree));

	item =
		(GtkTreeItem *) gtk_ctree_insert_node(GTK_CTREE(tree), NULL, NULL,
											  &itemtext1, 5, NULL, NULL, NULL,
											  NULL, 0, 1);
	gtk_widget_show(GTK_WIDGET(item));

	dbitem =
		GTK_WIDGET(gtk_ctree_insert_node
				   (GTK_CTREE(tree), GTK_CTREE_NODE(item), NULL, &dbname1[0],
					5, NULL, NULL, NULL, NULL, 1, 1));
	gtk_widget_show(GTK_WIDGET(dbitem));

	dbitem =
		GTK_WIDGET(gtk_ctree_insert_node
				   (GTK_CTREE(tree), GTK_CTREE_NODE(item),
					GTK_CTREE_NODE(dbitem), &dbname1[1], 5, NULL, NULL, NULL,
					NULL, 1, 1));
	gtk_widget_show(GTK_WIDGET(dbitem));

	item =
		GTK_TREE_ITEM(gtk_ctree_insert_node
					  (GTK_CTREE(tree), NULL, NULL, &itemtext2, 5, NULL, NULL,
					   NULL, NULL, 0, 1));
	gtk_widget_show(GTK_WIDGET(item));

	label = gtk_label_new("OK");
	gtk_widget_show(GTK_WIDGET(label));

	gtk_container_add(GTK_CONTAINER(item), label);

	return tree;
}

#endif
