/** \file guiapi.c	Interface to menu module.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup gui	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"
#include <stdlib.h>
#include <string.h>

#include "buf.h"
#include "db_connect_dlg.h"
#include "dbapi.h"
#include "dialogs.h"
#include "gui.h"
#include "guiapi.h"
#include "queries.h"
#include "status.h"
#include "tables.h"
#include "utils.h"
#include "workspace.h"

/** Add a new blank query.	*/
void gui_add_query(void) {
	query_add();
}

/** Returns true if the connection is good.
 * \return 1 if the connection is good, 0 otherwise.	*/
int gui_connection_ok(void)
{
	if (gui_get_connection() != NULL)
		return 1;
	else
		return 0;
}

/** Delete all open queries.	*/
void gui_delete_all_queries(void)
{
	int i, nb = nbQueries;

	/* Used to prevent bugs... */
	gtk_signal_disconnect(GTK_OBJECT(queryNotebook), queryNotebookPageChangeHandler);
	queryNotebookPageChangeHandler = 0;

	for (i = 0; i < nb; i++)
		query_delete_at_position(0);

	lastQuery = 0;
	nbQueries = 0;
}

/** Disconnects the GUI's database connection.	*/
void gui_disconnect(void) {
	DBConnection *conn = gui_get_connection();
	if (conn == NULL) {
		g_warning("Can't disconnect. Already disconnected.");
		return;
	}
	tables_connection_closed();
	gui_delete_all_queries();

	/* Resets the queries counter */
	queries_init();
	db_disconnect(conn, NULL);
	guiConnection = NULL;
}

/** Get the current connection.
 * \return The connection pointer.	*/
DBConnection *gui_get_connection(void)
{
	return guiConnection;
}

/** Get the URL for the current connection.
 * \return The url string. Don't free this.	*/
char *gui_get_connection_url(void)
{
	DBConnector *connector;
	DBConnection *conn = gui_get_connection();
	if (conn == NULL)
		return NULL;

	connector = conn->connector;
	if (connector == NULL)
		return NULL;

	return connector->form_to_url(connector->db_frm);
}

/** Find the number of rows in the current result.
 * \return The number of rows.	*/
int gui_get_query_count(void) {
	return query_count();
}

/** Find the number of columns in the current result.
 * \return The number of columns.	*/
int gui_get_result_columns()
{
	void *result;
	DBConnection *conn;

	/* Find the number of rows in the result.	*/
	conn = gui_get_connection();
	if (conn == NULL)
		return 0;

	result = query_get_result(query_get_current());
	if (result == NULL)
		return 0;

	if (conn->DBget_query_status(conn, result) != DB_QUERY_RECORDS_OK)
		return 0;

	return conn->DBget_field_number(conn, result);
}

/** Lookup a field text given a column and row.
 * \param row	The field row
 * \param column	The field column
 * \return	The field value as a string.	*/
char *gui_get_result_field(int row, int column)
{
	DBConnection *conn = gui_get_connection();
	void *result = query_get_current_result();
	int nb_tuples, nb_cols;
	int stat;

	stat = conn->DBget_query_status(conn, result);
	if (stat == DB_QUERY_RECORDS_OK) {
		nb_tuples = conn->DBget_record_number(conn, result);
		nb_cols = conn->DBget_field_number(conn, result);

		if ((row >= nb_tuples) || (column >= nb_cols))
			return "";

		return conn->DBget_field_value(conn, result, row, column);
	}
	else
		return "";
}

/** Find the number of rows in the current result.
 * \return The number of rows.	*/
int gui_get_result_rows(void)
{
	void *result;
	DBConnection *conn;

	/* Find the number of rows in the result.	*/
	conn = gui_get_connection();
	if (conn == NULL)
		return 0;

	result = query_get_result(query_get_current());
	if (result == NULL)
		return 0;

	if (conn->DBget_query_status(conn, result) != DB_QUERY_RECORDS_OK)
		return 0;

	return conn->DBget_record_number(conn, result);
}

/** Initialise the GUI.
 * \param argc	The argc parameter passed to main.
 * \param argv	The argv parameter passed to main.	*/
void gui_init(int argc, char **argv) {
	add_pixmap_directory(PACKAGE_PIXMAPS_DIR);

	/* Window / GTK initialisation */
	gtk_init(&argc, &argv);
	gtk_rc_parse(get_rcname());

	mainWindow = create_mainWindow();

	gtk_widget_show(mainWindow);
	populate_mainWindow();
	setup_widget_vars();

	db_connect_dialog_create();

	/* Activate/deactivate the appropriate widgets.	*/
	update_button_sensitivity();
}

/** Load the current query from a file.
 * \param filename	The file to load.	*/
void gui_load_query_from_file(char *filename)
{
	char *name;
	ABuf *buf = buf_new(256);
	FILE *fp;
	GtkWidget *cq = query_get_current();
	QueryData *qd = gtk_object_get_data(GTK_OBJECT(cq), QDATA_KEY);

	query_clear();

	gtk_text_freeze(GTK_TEXT(cq));
	fp = fopen(filename, "r");
	if (fp != NULL) {
		while (fgets(buf->b_dat, buf->b_len, fp) != NULL)
			gtk_text_insert(GTK_TEXT(cq), NULL, NULL, NULL, buf->b_dat, -1);

		fclose(fp);
		querydata_set_filename(qd, filename);
		querydata_set_unchanged(qd);
		name = get_short_filename(filename);
		if (name == NULL)
			name = "Untitled";
	
		gui_set_query_name(name);
	}
	gtk_text_thaw(GTK_TEXT(cq));
	buf_free(buf);
}

/** Run the GUI.
 * \return The return code to be used on exit.	*/
int gui_run(void) {
	/* Load the workspace */
	gui_get_workspace();

	running++;
	gtk_main();
	running--;
	return 0;
}

/** Send the query to the database.	*/
void gui_send_query() {
#define ERROREXEC_S	" Error executing query \n %s "
	GtkWidget *current_query = query_get_current();
	char *cquery = gtk_editable_get_chars(GTK_EDITABLE(current_query), 0, -1);
	void *result =
		(void *) gtk_object_get_data(GTK_OBJECT(current_query), RESULT_KEY);
	char *status =
		(char *) gtk_object_get_data(GTK_OBJECT(current_query), STATUS_KEY);
	DBConnection *conn = gui_get_connection();

	if (result != NULL)
		conn->DBclear_result(conn, result);

	/* Clear old error message. */
	results_display(NULL, QUERY_FAKE,
					gtk_notebook_current_page(GTK_NOTEBOOK(queryNotebook)));

	status_push("Query in progress...");

	result = conn->DBexecute_query(conn, cquery);
	gtk_object_set_data(GTK_OBJECT(current_query), RESULT_KEY, result);
	if ((conn->DBget_query_status(conn, result) != DB_QUERY_RECORDS_OK) &&
		(conn->DBget_query_status(conn, result) != DB_QUERY_COMMAND_OK) &&
		(conn->DBget_query_status(conn, result) != DB_QUERY_EMPTY)) {
		GtkWidget *dialog;
		char *msg = conn->DBget_error_message(conn);
		ABuf *buf = buf_new(strlen(ERROREXEC_S) + strlen(msg));
		if (!buf) {
			g_error("SendQuery() could not allocate memory for a buffer\n");
			return;
		}
		snprintf(buf->b_dat, buf->b_len, ERROREXEC_S, msg);
		dialog =
			create_standard_dialog("Error executing query", buf->b_dat, 1,
								   NULL, "Ok", NULL, NULL, NULL, NULL);
		gtk_widget_show(dialog);
		results_display((void *) buf->b_dat, QUERY_ERROR,
						gtk_notebook_current_page(GTK_NOTEBOOK
												  (queryNotebook)));

		sprintf(status, "Error in query");
		buf_free(buf);
	}
	else {
		int nbf = conn->DBget_field_number(conn, result);
		int nbr = conn->DBget_record_number(conn, result);

		results_display(result, QUERY_NEW,
						gtk_notebook_current_page(GTK_NOTEBOOK
												  (queryNotebook)));
		sprintf(status, "Query results : %d field%sand %d record%s", nbf,
				(nbf != 1 ? "s " : " "), nbr, (nbr != 1 ? "s" : ""));
	}

	g_free(cquery);

	status_pop();

	if (status != NULL) {
		if (queryStatus == 1)
			status_pop();
		status_push(status);
		queryStatus = 1;
	}
	else
		queryStatus = 0;
}

/** Set the currently active connection.
 * \param	con	The connection pointer to set.	*/
void gui_set_connection(DBConnection * con)
{
	char buf[500];
	char *s = NULL;

	if (guiConnection != NULL)
		gui_disconnect();

	guiConnection = con;

	tables_new_connection();
	gui_add_query();
	update_button_sensitivity();
	if (con != NULL)
		s = con->DBget_db_name(con);

	if (!s)
		s = "";

	snprintf(buf, sizeof(buf)-1, "GtkSQL: %s", s);
	buf[sizeof(buf)-1] = '\0';
	gtk_window_set_title(GTK_WINDOW(mainWindow), buf);
}

/** Set the name of the current query notebook page. If name is NULL or an empty string,
 * the call is ignored. Otherwise, the name is set as specified.
 * \param name	The new name.	*/
void gui_set_query_name(gchar *name) {
	int position = gtk_notebook_current_page(GTK_NOTEBOOK(queryNotebook));
	query_set_name(position, name);
}

/** Set the text for the currently active query.
 * \param txt	The text to set.	*/
void gui_set_query_text(char *txt)
{
	GtkWidget *cq = query_get_current();
	QueryData *qd = gtk_object_get_data(GTK_OBJECT(cq), QDATA_KEY);

	query_clear();

	/* We get the file name */
	gtk_text_freeze(GTK_TEXT(cq));
	gtk_text_insert(GTK_TEXT(cq), NULL, NULL, NULL, txt, -1);
	querydata_set_unchanged(qd);
	gtk_text_thaw(GTK_TEXT(cq));
}
