/** \file luaif.c	LUA engine handling functions.
 * \author Darryl Luff
 * \ingroup lua	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002-2003  Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"
#ifdef USE_LUA
#include <gtk/gtk.h>

#include <lua.h>
#include <lualib.h>

#include "guiapi.h"
#endif

/** Test if LUA interface is supported.
 * \return 1 if supported, 0 otherwise.	*/
int lua_supported(void) {
#ifdef USE_LUA
	return 1;
#else
	return 0;
#endif
}

/* Only include the rest if LUA is enabled.	*/

#ifdef USE_LUA

/** LUA state pointer.	*/
static lua_State *lua = NULL;

/* ## LUA interface calls	## */

/** Handle the addQuery(text) call. LUA usage "addQuery("querytext")"
 * \param  luaVM	The LUA state variable.
 */
int lic_addQuery(lua_State* luaVM) {
	int args = lua_gettop(luaVM);

	if ((args < 1) || (args > 2)) {
		printf("USAGE: addQuery([\"Title\", ] \"querytext\")\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}
	if (gui_connection_ok()) {
		/* Create a new query	*/
		gui_add_query();

		/* Set the text	*/
		gui_set_query_text((gchar *)lua_tostring(lua, -1));
		if (args == 2)
			gui_set_query_name((gchar *)lua_tostring(lua, -2));
	}
	lua_pushnumber(luaVM, 0);
	return 1;
}

/** Handle the columnCount() call. LUA usage "n = columnCount()"
 * \param  luaVM	The LUA state variable.	*/
int lic_columnCount(lua_State* luaVM) {
	int count;

	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 0) {
		g_warning("USAGE: columnCount()\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	/* Resize the window.	*/
	count = gui_get_result_columns();

	lua_pushnumber(luaVM, count);
	return 1;
}

/** Handle the connect(url) call. LUA usage: "connect(url)".
 * 'url' is of the form "mysql://user:pass@hostname/db=fred"
 * \param  luaVM	The LUA state variable.
 */
int lic_connect(lua_State* luaVM) {
	DBConnection *conn;

	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 1) {
		g_warning("USAGE: connect(\"url\")");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	/* Make a new connection.	*/
	conn = db_connect_url((gchar *)lua_tostring(lua, -1));
	gui_set_connection(conn);

	/* Close the default query.	*/
	gui_delete_all_queries();

	lua_pushnumber(luaVM, 0);
	return 1;
}

/** Handle the addQuery(text) call. LUA usage: "deleteQueries()"
 * \param  luaVM	The LUA state variable.
 */
int lic_deleteQueries(lua_State* luaVM) {
	/* Delete all queries.	*/

	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 0) {
		g_warning("USAGE: deleteQueries()\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	gui_delete_all_queries();

	lua_pushnumber(luaVM, 0);
	return 1;
}

/** Disconnect error callback.	*/
void disconnect_error(char *title, char *msg) {
	printf("%s\n%s\n\n", title, msg);
}

/** Handle the disconnect() call. LUA usage "disconnect()"
 * \param  luaVM	The LUA state variable.
 */
int lic_disconnect(lua_State* luaVM) {
	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 0) {
		g_warning("USAGE: disconnect()\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	if (gui_connection_ok()) {
		gui_disconnect();
	}
	lua_pushnumber(luaVM, 0);
	return 1;
}

/** Handle the exit(text) call. LUA usage: "exit()"
 * \param  luaVM	The LUA state variable.	*/
int lic_exit(lua_State* luaVM) {
	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 0) {
		g_warning("USAGE: exit()\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	/* Exit the application.	*/
	if (running != 0)
		gtk_main_quit();
	else
		exit(0);

	lua_pushnumber(luaVM, 0);
	return 1;
}

/** Handle the getField() call. LUA usage "field = getField(row, column)"
 * \param  luaVM	The LUA state variable.	*/
int lic_getField(lua_State* luaVM) {
	int row;
	int column;
	char *s;

	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 2) {
		g_warning("USAGE: getField(row, column)\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	/* Resize the window.	*/
	row = lua_tonumber(luaVM, 1);
	column = lua_tonumber(luaVM, 2);
	s = gui_get_result_field(row, column);
	lua_pushstring(lua, s);
	return 1;
}

/** Handle the exit(text) call. LUA usage: "exit()"
 * \param  luaVM	The LUA state variable.	*/
int lic_help(lua_State* luaVM) {
	/* Show some simple help.	*/
	printf("GtkSQL scripting commands:\n");
	printf("addQuery([\"title\",]\"query text\")\n\tAdd a query containing the given.\n");
	printf("\ttext and (optional) title.\n");
	printf("columnCount()\n\tReturn the number of columns in the query result.\n");
	printf("connect(\"connectionurl\")\n\tConnect to the given URL.\n");
	printf("disconnect()\n\tDisconnects if a database is connected\n");
	printf("deleteQueries()\n\tClose all open queries.\n");
	printf("exit()\n\tExit the application.\n");
	printf("getField(row, column)\n\tReturn the field text for the row/column specified\n");
	printf("help()\n\tThis text.\n");
	printf("loadQuery(\"filename\")\n\tOpen a new query from the given file.\n");
	printf("moveWindow(x, y)\n\tMove the top-left corner of the window to the given location.\n");
	printf("rowCount()\n\tReturns the number of rows in the current result.\n");
	printf("runQuery(\"query\")\n\tRun either the given query, or the current query\nif no query is passed here.\n");
	printf("sizeWindow(width, height)\n\tSet the window to the given width and height.\n");
	printf("\n\nNOTE: All standard LUA functions are also available.\n\n");

	lua_pushnumber(luaVM, 0);
	return 1;
}

/** Handle the loadQuery(text) call. LUA usage "loadQuery("filename")"
 * \param  luaVM	The LUA state variable.
 */
int lic_loadQuery(lua_State* luaVM) {
	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 1) {
		g_warning("USAGE: loadQuery(\"filename\")\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	/* Create a new query	*/
	gui_add_query();

	/* Load the file	*/
	gui_load_query_from_file((char *)lua_tostring(lua, -1));

	lua_pushnumber(luaVM, 0);
	return 1;
}

/** Handle the moveWindow(x, y) call. LUA usage "moveWindow(x, y)"
 * \param  luaVM	The LUA state variable.
 */
int lic_moveWindow(lua_State* luaVM) {
	int x;
	int y;

	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 2) {
		g_warning("USAGE: moveWindow(x, y)\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	x = lua_tonumber(luaVM, 1);
	y = lua_tonumber(luaVM, 2);

	/* Resize the window.	*/
	gtk_widget_set_uposition(GTK_WIDGET(mainWindow), x, y);

	lua_pushnumber(luaVM, 0);
	return 1;
}

/** Handle the rowCount() call. LUA usage "rows = rowCount()"
 * \param  luaVM	The LUA state variable.	*/
int lic_rowCount(lua_State* luaVM) {
	int count;

	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 0) {
		g_warning("USAGE: recordCount()\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	count = gui_get_result_rows();

	lua_pushnumber(luaVM, count);
	return 1;
}

/** Handle the runQuery() call. LUA usage "runQuery("query text")". The query text is optional.
 * if not present, the current query will be run.
 * \param  luaVM	The LUA state variable.	*/
int lic_runQuery(lua_State* luaVM) {
	int args = lua_gettop(luaVM);

	if (args > 1) {
		g_warning("USAGE: runQuery(query)\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	if (gui_connection_ok()) {
		if (args == 1) {
			/* Query specified.	*/
			if (gui_get_query_count() == 0)
				gui_add_query();

			/* Set the text	*/
			gui_set_query_text((gchar *)lua_tostring(lua, -1));
		}
		gui_send_query();
		lua_pushnumber(luaVM, 0);
	}
	else
		lua_pushnumber(luaVM, 1);
	return 1;
}

/** Handle the sizeWindow() call. LUA usage "sizeWindow(width, height)"
 * \param  luaVM	The LUA state variable.	*/
int lic_sizeWindow(lua_State* luaVM) {
	int width;
	int height;

	/* Get the arguments.	*/
	if (lua_gettop(luaVM) != 2) {
		g_warning("USAGE: sizeWindow(widgth, height)\n");
		lua_pushnumber(luaVM, 1);
		return 1;
	}

	width = lua_tonumber(luaVM, 1);
	height = lua_tonumber(luaVM, 2);

	/* Resize the window.	*/
	gtk_widget_set_usize(GTK_WIDGET(mainWindow), width, height);

	lua_pushnumber(luaVM, 0);
	return 1;
}

/* ## Interface routines	## */

void luaif_init(void) {
	if (lua != NULL)
		return;

	lua = lua_open();
	if (lua == NULL) {
		g_warning("Could not initialise scripting engine.\n");
		return;
	}

	luaopen_base(lua);
	lua_register(lua, "addQuery", lic_addQuery);
	lua_register(lua, "columnCount", lic_columnCount);
	lua_register(lua, "connect", lic_connect);
	lua_register(lua, "deleteQueries", lic_deleteQueries);
	lua_register(lua, "disconnect", lic_disconnect);
	lua_register(lua, "exit", lic_exit);
	lua_register(lua, "getField", lic_getField);
	lua_register(lua, "help", lic_help);
	lua_register(lua, "loadQuery", lic_loadQuery);
	lua_register(lua, "moveWindow", lic_moveWindow);
	lua_register(lua, "sizeWindow", lic_sizeWindow);
	lua_register(lua, "rowCount", lic_rowCount);
	lua_register(lua, "runQuery", lic_runQuery);
}

/** \param fn	The filename for the script.	*/
void luaif_runfile(char *fn) {
	if (lua == NULL)
		return;

	lua_dofile(lua, fn);
}

/** \param s	The script string to run.	*/
void luaif_runstring(char *s) {
	if ((lua == NULL) || (s == NULL))
		return;

	lua_dostring(lua, s);
}

void luaif_shutdown(void) {
	if (lua == NULL)
		return;

	lua_close(lua);
}

#endif
