/** \file queries.c	Functions to handle queries.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup gui	*/
/*
 * GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 1998-2003  Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of tdo_he License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gdk/gdkkeysyms.h>
#include <string.h>

#include "buf.h"
#include "callbacks.h"
#include "drag.h"
#include "export.h"
#include "guiapi.h"
#include "menus.h"
#include "queries.h"
#include "query_import_dlg.h"
#include "query_save_dlg.h"
#include "status.h"
#include "tables.h"
#include "utils.h"

/* Keys for widget data.	*/
#define POSITION_KEY "Position"

#define STATUS_SIZE 64

#define IS_SEPARATOR(c) ((c == ' ')  || (c == '.')  || (c == '\r') || \
			 (c == '\n') || (c == '\t') || (c == ',')  || \
			 (c == '(')  || (c == '('))

/* Popup menu	*/
static GtkWidget *queryPopup = NULL;	/**< Popup menu for the query pane.	*/
static GtkWidget *resultPopup = NULL;	/**< Popup menu for the result pane.	*/

/* Variables	*/
static Keyword *lastCompletion = NULL;
static int lastIndex = 0;
/** The index of the last query entry.	*/
int lastQuery = 0;
static char *lastStart = NULL;
int nbQueries = 0;			/**< Number of displayed queries.	*/
int queryStatus = 0;
guint queryNotebookPageChangeHandler = 0;

/* Query pane popup	*/
GtkWidget *queryPopupClearItem;		/**< Clear Query menu item.	*/
GtkWidget *queryPopupCloseItem;		/**< Close Query menu item.	*/
GtkWidget *queryPopupImportItem;	/**< Import Query menu item.	*/
GtkWidget *queryPopupNewItem;		/**< New Query menu item.	*/
GtkWidget *queryPopupSaveItem;		/**< Query Save menu item.	*/
GtkWidget *queryPopupSaveAsItem;	/**< Query Save As menu item.	*/
GtkWidget *queryPopupSendItem;		/**< Send Query menu item.	*/

/** Definition of query popup menu.	*/
SubMenu queryPopupTemplate[] = {
	{"Send", on_querySend_activate, NULL, &queryPopupSendItem, NORM_BUTTON},
	{"New", on_queryNew_activate, NULL, &queryPopupNewItem, NORM_BUTTON},
	{"Clear", on_queryClear_activate, NULL, &queryPopupClearItem, NORM_BUTTON},
	{"Import ...", on_queryImport_activate, NULL, &queryPopupImportItem, NORM_BUTTON},
	{"Save", on_querySave_activate, NULL, &queryPopupSaveItem, NORM_BUTTON},
	{"Save As ...", on_querySaveAs_activate, NULL, &queryPopupSaveAsItem, NORM_BUTTON},
	{"Close", on_queryClose_activate, NULL, &queryPopupCloseItem, NORM_BUTTON},
	{NULL, NULL, NULL, NULL, NORM_BUTTON}
};

/* Event response routines	*/
static gint changed_event(GtkWidget * widget, gpointer data);
static gint key_press_event(GtkWidget * widget, GdkEventKey * event,
							gpointer data);

/* Dialog callbacks	*/
gboolean on_query_button_press_event(GtkWidget * widget,
									 GdkEventButton * event,
									 gpointer user_data);

gboolean on_result_button_press_event(GtkWidget * widget,
									  GdkEventButton * event,
									  gpointer user_data);

/** Definition of results pane popup menu.	*/
GtkWidget *resultPopupExportItem = NULL;
SubMenu resultPopupMenu[] = {
	{"Export...", do_export_dialog, NULL, &resultPopupExportItem, NORM_BUTTON},
	{NULL, NULL, NULL, NULL, NORM_BUTTON}
};

/** Set the name of the query notebook page. If name is NULL or an empty string,
 * the call is ignored. Otherwise, the name is set as specified.
 * \param position	The query index number. Zero-based.
 * \param name	The new name.	*/
void query_set_name(int position, char *name) {
	if ((name != NULL) && (strlen(name) != 0)) {
		GtkWidget *page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(queryNotebook), position);
		gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(queryNotebook), page, name);
		gtk_notebook_set_menu_label_text(GTK_NOTEBOOK(queryNotebook), page, name);
	}
}

/** Handle the keyword completion.
 * \param widget	???.
 * \param data	The text widget.
 * \return	???	*/
static gint changed_event(GtkWidget * widget, gpointer data)
{
	GtkWidget *text = GTK_WIDGET(data);
	gint index;
	DBConnection *conn = gui_get_connection();
	QueryData *qd;

	index = gtk_text_get_point(GTK_TEXT(text));

	/* This is in case of a cursor movement without any changes */
	if (index != lastIndex + 1) {
		lastCompletion = NULL;
		g_free(lastStart);
		lastStart = NULL;
	}

	/* Mark the query as changed.	*/
	qd = query_get_querydata(text);
	if (qd) {
		querydata_set_changed(qd, 1);
		update_button_sensitivity();
	}

	switch (GTK_TEXT_INDEX(GTK_TEXT(text), index - 1)) {
	case '\t':{					   /* Tabulation */
			int lstart_length, linserted_length, word_start;
			Keyword *cur;

			gtk_text_freeze(GTK_TEXT(text));

			/* First, we test if it the first time we press on TAB */
			if (lastCompletion != NULL) {
				/* Not the first time -> we use the same "start letters" as
				 * before */
				cur = lastCompletion->next;
				linserted_length = strlen(lastCompletion->key);
			}
			else {
				/* We first search for the characters to complete */
				word_start = index - 2;
				while (word_start >= 0) {
					char c = GTK_TEXT_INDEX(GTK_TEXT(text), word_start);
					if (IS_SEPARATOR(c))
						break;
					word_start--;
				}
				word_start++;
				/* Now, we get the characters and compare them to the
				 * keywords */
				lastStart =
					gtk_editable_get_chars(GTK_EDITABLE(text), word_start,
										   index - 1);
				linserted_length = -1;

				/* We search the whole list of keywords */
				cur = conn->keylist;
			}
			/* We search for a keyword starting with lastStart */
			if ((lastStart != NULL)
				&& ((lstart_length = strlen(lastStart)) > 0)) {
				if (linserted_length < 0)
					linserted_length = lstart_length;

				lastCompletion = NULL;
				while (cur != NULL) {
					if (!strncasecmp(lastStart, cur->key, lstart_length)) {
						lastCompletion = cur;
						break;
					}

					cur = cur->next;
				}
			}

			/* Now we change the text */
			if (linserted_length < 0)
				linserted_length = 0;

			gtk_text_set_point(GTK_TEXT(text), index - 1 - linserted_length);
			if (lastCompletion != NULL) {
				/* We insert the new keyword */
				gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, cur->key,
								-1);
				gtk_text_set_point(GTK_TEXT(text), index + strlen(cur->key));
				lastIndex = index - 1 - linserted_length + strlen(cur->key);
			}
			else {
				if (lastStart != NULL) {
					/* Or the original letters */
					gdk_beep();
					gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL,
									lastStart, -1);
					gtk_text_set_point(GTK_TEXT(text),
									   index + strlen(lastStart));
					lastIndex =
						index - 1 - linserted_length + strlen(lastStart);
				}
				else {
					lastIndex = index - 1 - linserted_length;
				}
			}
			gtk_text_backward_delete(GTK_TEXT(text), linserted_length + 1);

			gtk_text_thaw(GTK_TEXT(text));
			break;
		}

	default:
		lastCompletion = NULL;
		g_free(lastStart);
		lastStart = NULL;
		lastIndex = index;
		break;
	}

	/* Now, syntax coloring */
	return (FALSE);
}

/** Clear the current query.	*/
void query_clear(void)
{
	GtkWidget *query = query_get_current();
	gtk_editable_delete_text(GTK_EDITABLE(query), 0, -1);
}

/** Create the popup menu for the result pane.
 * \return The result context menu.	*/
GtkWidget *result_popup_create(void)
{
	return create_menu(resultPopupMenu);
}

/** Delete the selected query. The position index is zero-based.
 * \param position	The index of the query to delete.
 * \return The number of queries remaining.	*/
int query_delete_at_position(int position)
{
	GtkWidget *query;
	void *result;
	gchar *status;
	DBConnection *conn = gui_get_connection();
	/* QueryData *qd;	*/
	GtkWidget *scroller;

	scroller = gtk_notebook_get_nth_page(GTK_NOTEBOOK(queryNotebook), position);
	query = query_get_no(position);

	/* Does it need to be saved?	*/
	/* auto-save disabled.
	qd = query_get_querydata(query);
	if (qd != NULL) {
		if (qd->changed)
			query_save(query);	Autosave disabled.
	}
	*/

	/* Free the data items.	*/
	result = (void *)gtk_object_get_data(GTK_OBJECT(query), RESULT_KEY);
	status = (gchar *)gtk_object_get_data(GTK_OBJECT(query), STATUS_KEY);

	if (status != NULL)
		g_free(status);

	if (result != NULL)
		conn->DBclear_result(conn, result);

	if (position > 0)
		gtk_widget_grab_focus(gtk_notebook_get_nth_page
							  (GTK_NOTEBOOK(queryNotebook), position - 1));
	else if (nbQueries > 1)
		gtk_widget_grab_focus(gtk_notebook_get_nth_page
							  (GTK_NOTEBOOK(queryNotebook), position + 1));

	gtk_notebook_remove_page(GTK_NOTEBOOK(queryNotebook), position);
	gtk_notebook_remove_page(GTK_NOTEBOOK(resultNotebook), position);
	nbQueries--;

	return nbQueries;
}

/** Display the results in the result pane.
 * \param result	Depends on 'why' param. For good queries, points to the database result structure.
 * For QUERY_ERROR, is an error message.
 * \param why	Why are the results being displayed? Can be:
 *  - QUERY_FAKE	A new label is set as the current page result widget.
 *  - QUERY_SWITCH	The result notepad is switched to the appropriate page.
 *  - QUERY_ERROR	A new label is set as the current page result widget.
 *  - Otherwise, the result widget is set to a CList containing the query results inside a scrolled window.
 * \param position	The result notebook page number to set the result on. If < 0,
 * a new page is appended.	*/
void results_display(void *result, int why, int position)
{
	DBConnection *conn = gui_get_connection();
	GtkWidget *result_widget = NULL;

	if (why == QUERY_FAKE) {		   /* Add a dummy message */
		result_widget = gtk_label_new("No query.");
		gtk_widget_show(result_widget);
		if (position >= 0) {
			gtk_notebook_remove_page(GTK_NOTEBOOK(resultNotebook), position);
			gtk_notebook_insert_page(GTK_NOTEBOOK(resultNotebook), result_widget, NULL, position);
		}
		else {
			gtk_notebook_append_page(GTK_NOTEBOOK(resultNotebook), result_widget, NULL);
			position = notebook_page_count(GTK_NOTEBOOK(resultNotebook))-1;
		}
		gtk_notebook_set_page(GTK_NOTEBOOK(resultNotebook), position);		
		gtk_widget_show(result_widget);
	}
	else if (why == QUERY_SWITCH) {		/* Just switch to the right page.	*/
		gtk_notebook_set_page(GTK_NOTEBOOK(resultNotebook), position);
	}
	else {								/* Both ERROR NEW need some extra processing.	*/
		if (why == QUERY_ERROR) {
			if (result == NULL)
				result_widget = gtk_label_new("Error in query.");
			else
				result_widget = gtk_label_new((char *) result);
		}
		else if (why == QUERY_NEW) {
			GtkWidget *result_clist = NULL;
			int nb_tuples, nb_cols, i, j;
			int maxwidth;
			int stat;
			char **col_titles;
			GdkFont *font;

			status_push("Displaying query results...");
			stat = conn->DBget_query_status(conn, result);
			if (stat == DB_QUERY_RECORDS_OK) {
				nb_tuples = conn->DBget_record_number(conn, result);
				nb_cols = conn->DBget_field_number(conn, result);

				/* Init the CList */
				col_titles = (char **) g_malloc(nb_cols * sizeof(char *));
				for (i = 0; i < nb_cols; i++)
					col_titles[i] = conn->DBget_field_name(conn, result, i);

				result_widget = gtk_scrolled_window_new(NULL, NULL);
				gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW
											   (result_widget),
											   GTK_POLICY_AUTOMATIC,
											   GTK_POLICY_AUTOMATIC);

				result_clist = gtk_clist_new_with_titles(nb_cols, col_titles);
				gtk_signal_connect(GTK_OBJECT(result_clist),
								   "button_press_event",
								   GTK_SIGNAL_FUNC
								   (on_result_button_press_event),
								   result_clist);
				gtk_container_add(GTK_CONTAINER(result_widget), result_clist);

				gtk_object_set_data(GTK_OBJECT(result_widget), RESULT_KEY, result_clist);

				font = gtk_widget_get_style(result_clist)->font;
				maxwidth = MAXCOLWID * gdk_char_width(font, 'X');
				for (i = 0; i < nb_cols; i++)
					gtk_clist_set_column_width(GTK_CLIST(result_clist), i,
											   gdk_string_width(font,
																col_titles
																[i]));

				/* Fill the CList */
				gtk_clist_freeze(GTK_CLIST(result_clist));
				for (i = 0; i < nb_tuples; i++) {
					for (j = 0; j < nb_cols; j++) {
						int width;

						col_titles[j] =
							conn->DBget_field_value(conn, result, i, j);

						/* take care of null fields */
						if (col_titles[j] == NULL)
							col_titles[j] = "";

						width = gdk_string_width(font, col_titles[j]);

						if (width > GTK_CLIST(result_clist)->column[j].width) {
							if (width > maxwidth)
								gtk_clist_set_column_width(GTK_CLIST
														   (result_clist), j,
														   maxwidth);
							else
								gtk_clist_set_column_width(GTK_CLIST
														   (result_clist), j,
														   width);
						}
					}
					gtk_clist_append(GTK_CLIST(result_clist), col_titles);
				}
				gtk_clist_thaw(GTK_CLIST(result_clist));
				g_free(col_titles);
				gtk_widget_show(result_clist);
			}
			else if (conn->DBget_query_status(conn, result) ==
					 DB_QUERY_COMMAND_OK)
				result_widget = gtk_label_new("Command ok.");
			else if (conn->DBget_query_status(conn, result) == DB_QUERY_EMPTY)
				result_widget = gtk_label_new("Empty query.");
			else
				result_widget = gtk_label_new("Error in query.");

			status_pop();
		}
		gtk_widget_show(result_widget);
		gtk_notebook_remove_page(GTK_NOTEBOOK(resultNotebook), position);
		gtk_notebook_insert_page(GTK_NOTEBOOK(resultNotebook), result_widget,
								 NULL, position);
		gtk_notebook_set_page(GTK_NOTEBOOK(resultNotebook), position);
	}
}

/** Return the query text widget selected by 'n'. 'n' is zero-based.
 * \return	The result object from the current query.	*/
void *query_get_current_result(void)
{
	return query_get_result(query_get_current());
}

/** Run the export dialog.
 * \param widget	The sending widget.
 * \param data	Unused.	*/
void do_export_dialog(GtkWidget * widget, gpointer data)
{
	result_export();
	update_button_sensitivity();
}

/** Event handler for key presses. Handle all the hot-keys.
 * \param widget	The sending widget.
 * \param event	The keypress event.
 * \param data	The text widget.	*/
static gint key_press_event(GtkWidget * widget, GdkEventKey * event,
							gpointer data)
{
	GtkWidget *text = GTK_WIDGET(data);

	if (event->state == GDK_MOD1_MASK) {
		/* We do not transmit to the Text Widget any of the Alt event */
		gtk_signal_emit_stop_by_name(GTK_OBJECT(data), "key_press_event");

		/* Alt - t = Display table */
		if (event->keyval == 't') {
			gint index, start, end, length;

			index = gtk_text_get_point(GTK_TEXT(text));
			length = gtk_text_get_length(GTK_TEXT(text));
			start = index;
			end = index;
			while ((start >= 0) &&
				   (!IS_SEPARATOR(GTK_TEXT_INDEX(GTK_TEXT(text), start))))
				start--;
			start++;
			while ((end < length) &&
				   (!IS_SEPARATOR(GTK_TEXT_INDEX(GTK_TEXT(text), end))))
				end++;

			display_table_named(gtk_editable_get_chars(GTK_EDITABLE(text),
													   start, end));

			return (TRUE);
		}
		/* Alt - n = New query */
		if (event->keyval == 'n') {
			query_add();
			return (TRUE);
		}
		/* Alt - d = Delete query */
		if (event->keyval == 'd') {
			if (nbQueries > 1) {
				query_close();
			}
			return (TRUE);
		}
		/* Alt - s = Send Query */
		if (event->keyval == 's') {
			gui_send_query();
			return (TRUE);
		}
		/* Alt - e = Export Query */
		if (event->keyval == 'e') {
			result_export();
			return (TRUE);
		}

		/* Alt arrows = switch between queries */
		if (event->keyval == GDK_Left) {
			int curpage;

			gtk_notebook_prev_page(GTK_NOTEBOOK(queryNotebook));

			curpage = gtk_notebook_current_page(GTK_NOTEBOOK(queryNotebook));
			gtk_widget_grab_focus(gtk_notebook_get_nth_page
								  (GTK_NOTEBOOK(queryNotebook), curpage));

			return (TRUE);
		}
		if (event->keyval == GDK_Right) {
			int curpage;

			gtk_notebook_next_page(GTK_NOTEBOOK(queryNotebook));

			curpage = gtk_notebook_current_page(GTK_NOTEBOOK(queryNotebook));
			gtk_widget_grab_focus(gtk_notebook_get_nth_page
								  (GTK_NOTEBOOK(queryNotebook), curpage));

			return (TRUE);
		}
	}
	return (FALSE);
}

/** Callback for query notebook page change.
 * \param notebook	The notebook widget.
 * \param page	The notebook page.
 * \param page_num	The page index.
 * \param data	Not used.	*/
void notebook_selection_callback(GtkNotebook * notebook, GtkNotebookPage * page, gint page_num, gpointer data)
{
	GList *children;
	GtkWidget *query;
	char *status;

	GtkWidget *scroller = page->child;
	if (scroller == NULL)
		return;

	children = gtk_container_children(GTK_CONTAINER(scroller));
	if (children == NULL)
		return;

	query = GTK_WIDGET(children->data);
	if (!query)
		return;

	status = (char *) gtk_object_get_data(GTK_OBJECT(query), STATUS_KEY);

	results_display(NULL, QUERY_SWITCH, page_num);

	update_button_sensitivity();
	if (status != NULL) {
		if (queryStatus == 1)
			status_pop();
		status_push(status);
		queryStatus = 1;
	}
	else
		queryStatus = 0;
}

/** Create the popup menu for the query pane.
 * \return The query context menu.	*/
static GtkWidget *query_popup_create(void)
{
	return create_menu(queryPopupTemplate);
}

/** Popup query menu on button click.
 * \param widget	The button.
 * \param event	The button event.
 * \param user_data	Not used.
 * \return	Boolean event status.	*/
gboolean on_query_button_press_event(GtkWidget * widget,
									 GdkEventButton * event,
									 gpointer user_data)
{
	if (event->button == 3) {
		if (queryPopup == NULL) {
			queryPopup = query_popup_create();
			if (queryPopup == NULL) {
				g_error("Couldnt create menu\n");
				return FALSE;
			}
		}
		gtk_menu_popup(GTK_MENU(queryPopup), NULL, NULL, NULL, NULL,
					   event->button, event->time);
		return TRUE;
	}
	return FALSE;
}

/** Popup result menu on button click.
 * \param widget	The button.
 * \param event	The button event.
 * \param user_data	Not used.	*/
gboolean on_result_button_press_event(GtkWidget * widget,
									  GdkEventButton * event,
									  gpointer user_data)
{
	if (event->button == 3) {
		if (resultPopup == NULL) {
			resultPopup = result_popup_create();
			if (resultPopup == NULL) {
				g_error("Couldnt create menu\n");
				return FALSE;
			}
		}
		gtk_menu_popup(GTK_MENU(resultPopup), NULL, NULL, NULL, NULL,
					   event->button, event->time);
		return TRUE;
	}
	return FALSE;
}

/** Return a copy of the text from the passed query.
 * \param query	The query widget.
 * \return A copy of the text. This must be freed with g_free when finished.	*/
char *query_get_text(GtkWidget * query)
{
	return gtk_editable_get_chars(GTK_EDITABLE(query), 0, -1);
}

void query_save_to_file(GtkWidget *query, char *fn) {
	FILE *fp;
	char *cquery;
	char *name;

	if ((query == NULL) || (fn == NULL)) {
		g_warning("query_save_to_file(%p, %p)\n", query, fn);
		return;
	}

	if (strlen(fn) < 1) {
		g_warning("query_save_to_file(). Ignoring null filename\n");
		return;
	}

	cquery = gtk_editable_get_chars(GTK_EDITABLE(query), 0, -1);

	fp = fopen(fn, "wb");
	if (fp != NULL) {
		fprintf(fp, "%s\n", cquery);
		fclose(fp);
		name = get_short_filename(fn);
		if (name == NULL)
			name = "Untitled";
	
		gui_set_query_name(name);
	}
}

/** \return The scrolling view containing the new query text widget.	*/
GtkWidget *query_add(void)
{
	GtkWidget *label;
	GtkWidget *newQuery;
	GtkWidget *scroller;
	QueryData *qd;
	ABuf *buf = buf_new(64);
	static GtkTargetEntry target = {
		"text/plain", 0, 0
	};
	char *status_buffer;

	/* Scroller */
	scroller = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroller),
								   GTK_POLICY_AUTOMATIC,
								   GTK_POLICY_AUTOMATIC);

	/* The query text zone */
	newQuery = gtk_text_new(NULL, NULL);

	gtk_text_set_editable(GTK_TEXT(newQuery), TRUE);
	gtk_signal_connect(GTK_OBJECT(newQuery), "changed",
					   GTK_SIGNAL_FUNC(changed_event), newQuery);
	gtk_signal_connect(GTK_OBJECT(newQuery), "key_press_event",
					   GTK_SIGNAL_FUNC(key_press_event), newQuery);
	gtk_signal_connect(GTK_OBJECT(newQuery), "button_press_event",
					   GTK_SIGNAL_FUNC(on_query_button_press_event),
					   newQuery);
	gtk_widget_show(newQuery);

	qd = querydata_new();
	gtk_object_set_data(GTK_OBJECT(newQuery), QDATA_KEY, qd);
	gtk_object_set_data(GTK_OBJECT(newQuery), RESULT_KEY, NULL);

	/* Setup to receive dropped text */
	gtk_signal_connect(GTK_OBJECT(newQuery), "drag_data_received",
					   GTK_SIGNAL_FUNC(receive_callback), NULL);
	gtk_drag_dest_set(GTK_WIDGET(newQuery),
					  GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_HIGHLIGHT |
					  GTK_DEST_DEFAULT_DROP, &target, 1, GDK_ACTION_COPY);

	/* Add the text field to the scroller */
	gtk_container_add(GTK_CONTAINER(scroller), newQuery);
	gtk_widget_show(scroller);

	/* Add the query to the notebook */
	snprintf(buf->b_dat, buf->b_len, "Query%d", ++lastQuery);
	label = gtk_label_new(buf->b_dat);
	gtk_widget_show(GTK_WIDGET(label));
	gtk_notebook_append_page(GTK_NOTEBOOK(queryNotebook), scroller, label);
	gtk_object_set_data(GTK_OBJECT(newQuery), POSITION_KEY, (void *)nbQueries);
	gtk_notebook_set_page(GTK_NOTEBOOK(queryNotebook), nbQueries);

	/* Add an empty result to the result notebook */
	results_display(NULL, QUERY_FAKE, -1);

	/* Add to the query a buffer to display a query status */
	status_buffer = (char *) malloc(STATUS_SIZE * sizeof(char));
	strcpy(status_buffer, "Ready");
	gtk_object_set_data(GTK_OBJECT(newQuery), STATUS_KEY,
						(void *) status_buffer);
	if (nbQueries == 0) {
		status_push(status_buffer);
		queryStatus = 1;
	}

	nbQueries++;
	buf_free(buf);

	if (queryNotebookPageChangeHandler == 0) {
		queryNotebookPageChangeHandler = gtk_signal_connect(GTK_OBJECT(queryNotebook), "switch_page",
						   GTK_SIGNAL_FUNC(notebook_selection_callback), NULL);
	}

	//results_display(NULL, QUERY_FAKE, nbQueries-1);
	update_button_sensitivity();
	return scroller;
}

/** Close the current query.
 * \return The number of queries remaining.	*/
int query_close()
{
	gint current_page;

	current_page = gtk_notebook_current_page(GTK_NOTEBOOK(queryNotebook));

	if (queryStatus == 1) {
		queryStatus = 0;
		status_pop();
	}
	return query_delete_at_position(current_page);
}

/** \return The current query text widget.	*/
GtkWidget *query_get_current()
{
	/* The notebook page is the scroller. */
	int cur_page = gtk_notebook_current_page(GTK_NOTEBOOK(queryNotebook));
	GtkWidget *scroller =
		gtk_notebook_get_nth_page(GTK_NOTEBOOK(queryNotebook), cur_page);
	GList *children;

	if (scroller == NULL)
		return NULL;

	children = gtk_container_children(GTK_CONTAINER(scroller));
	if (children == NULL)
		return NULL;

	return GTK_WIDGET(children->data);
}

/** Return the query text widget selected by 'n'. 'n' is zero-based.
 * \param n	The index of the entry to return.
 * \return	The selected query text widget.	*/
GtkWidget *query_get_no(int n)
{
	/* The notebook page is the scroller. */
	GtkWidget *scroller =
		gtk_notebook_get_nth_page(GTK_NOTEBOOK(queryNotebook), n);
	GList *children;

	if (scroller == NULL)
		return NULL;

	children = gtk_container_children(GTK_CONTAINER(scroller));
	if (children == NULL)
		return NULL;

	return GTK_WIDGET(children->data);
}

/** Return the text from the selected query.
 * \param n	The query to retrieve.
 * \return The query text.	*/
char *query_get_text_no(int n)
{
	GtkWidget *qry = query_get_no(n);
	if (!qry)
		return NULL;

	return query_get_text(qry);
}

/** \return The query name.	*/
char *query_get_current_name()
{
	GtkWidget *label;
	int position = gtk_notebook_current_page(GTK_NOTEBOOK(queryNotebook));
	char *qname;

	label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(queryNotebook),
									   gtk_notebook_get_nth_page(GTK_NOTEBOOK
																 (queryNotebook),
																 position));
	gtk_label_get(GTK_LABEL(label), &qname);

	return qname;
}

/** Get the query data for the passed query.
 * \param query	The query to access.
 * \return The QueryData pointer.	*/
QueryData *query_get_querydata(GtkWidget *query) {
	QueryData *qd = gtk_object_get_data(GTK_OBJECT(query), QDATA_KEY);
	if (qd == NULL)
		qd = querydata_new();

	return qd;
}

/** Return the query result data.
 * \param query	The query of interest.
 * \return A pointer to the result data.	*/
void *query_get_result(GtkWidget * query)
{
	if (query == NULL)
		return NULL;

	return (void *) gtk_object_get_data(GTK_OBJECT(query), RESULT_KEY);
}

/** Initialise the query counter.	*/
void queries_init()
{
	lastQuery = 0;
	if (queryNotebookPageChangeHandler == 0) {
		queryNotebookPageChangeHandler = gtk_signal_connect(GTK_OBJECT(queryNotebook), "switch_page",
						   GTK_SIGNAL_FUNC(notebook_selection_callback), NULL);
	}
}

/** Initialise the query notebook.
 * \return The query notebook widget.	*/
GtkWidget *query_pane_init()
{
	/* The query notebook */
	queryNotebook = gtk_notebook_new();
	gtk_widget_set_name(GTK_WIDGET(queryNotebook), "queries");
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(queryNotebook), GTK_POS_TOP);
	gtk_notebook_set_scrollable(GTK_NOTEBOOK(queryNotebook), TRUE);
	gtk_notebook_popup_enable(GTK_NOTEBOOK(queryNotebook));
	gtk_widget_set_usize(queryNotebook, 250, 130);

	queryNotebookPageChangeHandler =
		gtk_signal_connect(GTK_OBJECT(queryNotebook), "switch_page",
						   GTK_SIGNAL_FUNC(notebook_selection_callback),
						   NULL);
	gtk_widget_show(queryNotebook);
	return queryNotebook;
}

/** Initialise the results pane.
 * \return The container widget holding the pane.	*/
GtkWidget *result_pane_init()
{
	GtkWidget *container = gtk_vbox_new(FALSE, 0);

	resultNotebook = gtk_notebook_new();
	gtk_widget_set_name(GTK_WIDGET(resultNotebook), "results");
	gtk_notebook_popup_enable(GTK_NOTEBOOK(resultNotebook));
	gtk_box_pack_start(GTK_BOX(container), resultNotebook, TRUE, TRUE, 0);
	gtk_notebook_set_show_tabs(GTK_NOTEBOOK(resultNotebook), FALSE);

	gtk_widget_show(container);
	gtk_widget_show(resultNotebook);
	return container;
}

/** Import a query from a file.	*/
void query_import(void)
{
	query_import_dialog_show();
}

/** \return	The number of queries.	*/
int query_count()
{
	return nbQueries;
}

/** Save a query to disk.
 * \param qry	The query to save. If NULL, the current query is saved.	*/
void query_save(GtkWidget *qry)
{
	GtkWidget *query;
	QueryData *qd;

	if (qry == NULL)
		query = query_get_current();
	else
		query = qry;

	if (query == NULL)
		return;

	qd = query_get_querydata(query);
	if (! qd)
		return;

	if (qd->filename == NULL)
		query_save_as(NULL);
	else
		query_save_to_file(query, qd->filename);
}

/** Save the current query to disk.	*/
void query_save_as(GtkWidget *qry)
{
	query_save_dialog_show(qry);
}
