/** \file tables.c	Functions to handle the table information pane.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup gui	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2003 Lionel ULMER, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/** Table list interface.
 * This module handles all interaction with the table notebook.
 */

#include <string.h>
#include "callbacks.h"
#include "dialogs.h"
#include "drag.h"
#include "guiapi.h"
#include "menus.h"
#include "queries.h"
#include "tables.h"

/** The table notebook widget	*/
GtkWidget *tableNotebook;

/** Definition of table popup menu.	*/
GtkWidget *tablePopupDeleteQueryItem;	/**< Generate a DELETE query.	*/
GtkWidget *tablePopupInsertQueryItem;	/**< Generate a INSERT query.	*/
GtkWidget *tablePopupRefreshItem;		/**< Refresh the table listing.	*/
GtkWidget *tablePopupSelectQueryItem;	/**< Generate a SELECT query.	*/
GtkWidget *tablePopupUpdateQueryItem;	/**< Generate a UPDATE query.	*/

/** Template for table popup menu.	*/
SubMenu tablePopupTemplate[] = {
	{"<- SELECT Query", on_select_query_activate, NULL, &tablePopupSelectQueryItem, NORM_BUTTON},
	{"<- INSERT Query", on_insert_query_activate, NULL, &tablePopupInsertQueryItem, NORM_BUTTON},
	{"<- UPDATE Query", on_update_query_activate, NULL, &tablePopupUpdateQueryItem, NORM_BUTTON},
	{"<- DELETE Query", on_delete_query_activate, NULL, &tablePopupDeleteQueryItem, NORM_BUTTON},
	{"", NULL, NULL, NULL, NORM_BUTTON},
	{"Refresh", on_tableRefresh_activate, NULL, &tablePopupRefreshItem, NORM_BUTTON},
	{NULL, NULL, NULL, NULL, NORM_BUTTON}
};

/** A count of the tables in the widget list.	*/
static int tablesInList = 0;

/** The table widget container.	*/
static GtkWidget *container;

/** The popup table menu.	*/
static GtkWidget *tablePopup = NULL;

GtkWidget *tables_create_popup(void);
static GtkWidget *create_table_info(DBTableDef * tbd);
gboolean on_info_clist_button_press_event(GtkWidget * widget,
										  GdkEventButton * event,
										  gpointer user_data);

/** Selects the table with the specified name.
 * The named table is located in the table pane and 
 * displayed.
 */
int display_table_named(char *tablename)
{
	int i;

	for (i = 0; i < tablesInList; i++) {
		char *name;
		GtkWidget *label =
			gtk_notebook_get_tab_label(GTK_NOTEBOOK(tableNotebook),
									   gtk_notebook_get_nth_page(GTK_NOTEBOOK
																 (tableNotebook),
																 i));
		gtk_label_get(GTK_LABEL(label), &name);

		if (!strcmp(name, tablename)) {
			gtk_notebook_set_page(GTK_NOTEBOOK(tableNotebook), i);
			return 1;
		}
	}
	return 0;
}

/** Return the currently visible table pane.
 * The returned item is the GtkCList containing the table fields.	*/
GtkCList *get_current_table(void)
{
	/* GtkWidget *widget; */

	int cur_page = gtk_notebook_current_page(GTK_NOTEBOOK(tableNotebook));
	return GTK_CLIST(get_table_no(cur_page));
}

/** \return The name of the currently visible table pane.	*/
char *get_current_table_name(void)
{
	GtkWidget *page;
	GtkWidget *label;
	char *name;
	int curpage = gtk_notebook_current_page(GTK_NOTEBOOK(tableNotebook));
	page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(tableNotebook), curpage);
	if (!page)
		return NULL;

	label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(tableNotebook), page);
	if (!label)
		return NULL;

	gtk_label_get(GTK_LABEL(label), &name);
	return name;
}

/** \return A list of the selected table fields. The list is returned as a comma-separated text string.
 * If no fields are selected, returns "*".	*/
char *get_selected_table_fields(void)
{
	GtkCList *clist;
	GList *list;
	int row = 0;
	static char fields[2048] = "";
	char *field;
	int focus = -1;

	clist = (GtkCList *) get_current_table();
	if (!clist)
		return NULL;

	focus = clist->focus_row;
	gtk_clist_get_text(clist, focus, 0, &field);
	strcpy(fields, field);

	list = clist->selection;
	if (!list)
		strcpy(fields, "*");
	else {
		while (list) {
			row = (int) list->data;
			if (row != focus) {
				gtk_clist_get_text(clist, row, 0, &field);
				strcat(fields, ", ");
				strcat(fields, field);
			}
			list = list->next;
		}
	}
	return fields;
}

/** Returns the n'th table.	
 * \param	n	The table to return.
 * \return	A pointer to the clist that the table info is contained in.	*/
GtkCList *get_table_no(int n)
{
	GtkWidget *widget;
	GList *list;				/* child list */

	if (!(widget = gtk_notebook_get_nth_page(GTK_NOTEBOOK(tableNotebook), n)))
		return NULL;

	/* widget should be the scrolled window */
	if (!(list = gtk_container_children(GTK_CONTAINER(widget))))
		return NULL;

	if (!(widget = list[0].data))
		return NULL;

	/* widget is now the viewport */
	if (!(list = gtk_container_children(GTK_CONTAINER(widget))))
		return NULL;
	else
		widget = list[0].data;

	/* widget SHOULD now be the CList */
	return GTK_CLIST(widget);
}

/** Initialises the table pane.
 * \return The container widget to insert into the notebook.	*/
GtkWidget *init_tables_pane(void)
{
	container = gtk_vbox_new(FALSE, 0);

	tableNotebook = gtk_notebook_new();
	gtk_widget_set_name(GTK_WIDGET(tableNotebook), "tables");
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tableNotebook), GTK_POS_TOP);
	gtk_notebook_set_scrollable(GTK_NOTEBOOK(tableNotebook), TRUE);
	gtk_notebook_popup_enable(GTK_NOTEBOOK(tableNotebook));
	gtk_box_pack_start(GTK_BOX(container), tableNotebook, TRUE, TRUE, 0);

	gtk_widget_show(container);
	gtk_widget_show(tableNotebook);
	return container;
}

/** Removes all tables from the table notebook.
 * This is called when a database connection is closed.	*/
void tables_connection_closed(void)
{
	int i;
	DBConnection *conn = gui_get_connection();
	if (conn != NULL)
		conn->DBfree_table_def(conn);

	for (i = 0; i < tablesInList; i++) {
		gtk_notebook_remove_page(GTK_NOTEBOOK(tableNotebook), 0);
	}
}

/** Initialises table display for a new connection.
 * The list of tables is recalled and field displays created.	*/
void tables_new_connection(void)
{
	DBConnection *conn = gui_get_connection();
	if (conn == NULL)
		return;

	if (conn->DBget_table_def(conn) < 0) {
		GtkWidget *dialog = create_standard_dialog("Error retrieving table list", "Error retrieving table list", 1, NULL, "Ok", NULL, NULL, NULL, NULL);
		gtk_widget_show(dialog);
	}
	else {
		int i;
		tablesInList = conn->DBget_table_num(conn);

		for (i = 0; i < tablesInList; i++) {
			DBTableDef *tbd = conn->DBget_table(conn, i);
			GtkWidget *label;

			label = gtk_label_new(tbd->name);
			gtk_widget_show(label);
			gtk_notebook_append_page(GTK_NOTEBOOK(tableNotebook),
									 create_table_info(tbd), label);
		}
	}
}

/** Creates the popup menu for the table notebook.
 * The popup menu is stored in the tablePopup global variable.
 */
GtkWidget *tables_create_popup(void)
{
	return create_menu(tablePopupTemplate);
}

/** Creates the text for drag/drop from the table pane.
 * A text/plain collection of currectly selected field names is created.
 */
static GtkWidget *create_table_info(DBTableDef * tbd)
{
	GtkWidget *info_clist;
	GtkWidget *info_scroll;
	static GtkTargetEntry target = { "text/plain", 0, 0 };
	int i;
	char *col_titles[] = {
		"Field", "Type", "Length"
	};

	/* CList initialisation */
	info_scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(info_scroll),
								   GTK_POLICY_AUTOMATIC,
								   GTK_POLICY_AUTOMATIC);

	info_clist = gtk_clist_new_with_titles(3, col_titles);
	gtk_clist_set_selection_mode((GtkCList *) info_clist,
								 GTK_SELECTION_MULTIPLE);
	gtk_signal_connect(GTK_OBJECT(info_clist), "drag_data_get",
					   GTK_SIGNAL_FUNC(send_callback), NULL);
	gtk_signal_connect(GTK_OBJECT(info_clist), "button_press_event",
					   GTK_SIGNAL_FUNC(on_info_clist_button_press_event),
					   NULL);
	gtk_drag_source_set(info_clist, GDK_BUTTON1_MASK, &target, 1,
						GDK_ACTION_COPY);
	gtk_widget_show(info_clist);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(info_scroll),
										  info_clist);

	gtk_clist_set_column_width(GTK_CLIST(info_clist), 0, 130);
	gtk_clist_set_column_width(GTK_CLIST(info_clist), 1, 130);

	for (i = 0; i < tbd->field_num; i++) {
		char *line[3];

		line[0] = tbd->field_def[i].name;
		line[1] = tbd->field_def[i].type;
		line[2] = tbd->field_def[i].length;

		gtk_clist_append(GTK_CLIST(info_clist), line);
	}

	gtk_widget_show(info_scroll);
	return info_scroll;
}

/** 'Delete Query' action function.
 * Called from the toolbar and menus.
 */
void on_delete_query_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	char buff[2048] = "";
	GtkWidget *query = NULL;
	char *table = get_current_table_name();

	if (!table) {
		g_error("Could not get current table name\n");
		return;
	}

	snprintf(buff, sizeof(buff), "delete from %s\nwhere ...", table);

	query = query_get_current();
	if (query) {
		gtk_editable_delete_text(GTK_EDITABLE(query), 0, -1);
		gtk_text_insert(GTK_TEXT(query), NULL, NULL, NULL, buff, -1);
	}
}

/** 'click' event for table pane.
 * Pops up the popup menu on button 3 being pressed.
 */
gboolean on_info_clist_button_press_event(GtkWidget * widget,
										  GdkEventButton * event,
										  gpointer user_data)
{
	if (event->button == 3) {
		if (tablePopup == NULL) {
			tablePopup = tables_create_popup();
			if (tablePopup == NULL) {
				printf("Couldnt create menu\n");
				return FALSE;
			}
		}
		gtk_menu_popup(GTK_MENU(tablePopup), NULL, NULL, NULL, NULL,
					   event->button, event->time);
		return TRUE;
	}
	return FALSE;
}

/** 'Insert Query' action function.
 * Called by menu and toolbar actions.
 */
void on_insert_query_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	char buff[2048] = "";
	GtkWidget *query = NULL;
	char *table = get_current_table_name();
	char *fields = get_selected_table_fields();

	if (!fields) {
		g_error("Invalid fields\n");
		return;
	}

	if (!table) {
		g_error("Invalid tablename\n");
		return;
	}

	if (strcmp(fields, "*") == 0)
		snprintf(buff, sizeof(buff), "insert into %s\nvalues(...)\n", table);
	else
		snprintf(buff, sizeof(buff), "insert into %s (%s)\nvalues(...)\n",
				 table, fields);

	query = query_get_current();
	if (query) {
		gtk_editable_delete_text(GTK_EDITABLE(query), 0, -1);
		gtk_text_insert(GTK_TEXT(query), NULL, NULL, NULL, buff, -1);
	}
}

/** 'select query' action function.
 * Generates a default 'select' query from the currently selected 
 * table fields.	*/
void on_select_query_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	char buff[2048] = "";
	GtkWidget *query = NULL;
	char *table = get_current_table_name();
	char *fields = get_selected_table_fields();

	if (!fields) {
		g_error("Invalid fields\n");
		return;
	}

	if (!table) {
		g_error("Invalid tablename\n");
		return;
	}

	snprintf(buff, sizeof(buff), "select %s\nfrom %s\n", fields, table);

	query = query_get_current();
	if (query) {
		gtk_editable_delete_text(GTK_EDITABLE(query), 0, -1);
		gtk_text_insert(GTK_TEXT(query), NULL, NULL, NULL, buff, -1);
	}
}

/** 'update query' action function.
 * Generates a default 'update' query from the currently selected 
 * table fields.
 */
void on_update_query_activate(GtkMenuItem * menuitem, gpointer user_data)
{
	char buff[2048] = "";
	GtkWidget *query = NULL;
	char *table = get_current_table_name();

	if (!table) {
		g_error("Invalid tablename\n");
		return;
	}

	snprintf(buff, sizeof(buff), "update %s\nset ...\nwhere ...\n", table);

	query = query_get_current();
	if (query) {
		gtk_editable_delete_text(GTK_EDITABLE(query), 0, -1);
		gtk_text_insert(GTK_TEXT(query), NULL, NULL, NULL, buff, -1);
	}
}
