/** \file workspace.c	Workspace manager to load and save the
 * applications state.
 * \author Darryl Luff
 * \ingroup gui	*/
/* $Id: workspace.c,v 1.24 2003/08/31 09:47:50 djluff Exp $
 * 
 * GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 2002-2003 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Workspace handling. Hold implementation pending separation of db/interface 
 * code.
 *
 * Workspace format:
 * GTKSQL WORKSPACE
 * QUERY "select ..."
 * QUERY "insert into ..."
 * ...
 */

#include "../config.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "buf.h"
#include "guiapi.h"
#include "queries.h"
#include "luaif.h"
#include "workspace.h"

/** Character to map CR's in the SQL to.	*/
#define CR_CHAR '\\'

/* Function prototypes.	*/
void map_chars(char *sql, char c1, char c2);

/** Replace all of the 'c1' chars in txt to 'c2'.	*/
void map_chars(char *txt, char c1, char c2)
{
	char *c = txt;
	if ((!txt) || (strlen(txt) == 0))
		return;

	while (*c) {
		if (*c == c1)
			*c = c2;
		c++;
	}
}

/** Load the workspace from a file. Sets up the workspace variable, and loads the
 * queries, but does NOT disconnect the old database.
 * \param	fn	The filename to open.	*/
void ws_load_file(char *fn)
{
	if (fn == NULL)
		return;
#ifdef USE_LUA
	luaif_runfile(fn);
#else
	printf("WARNING: Workspace loading disabled. No scripting support.\n");
#endif
}

/** Save the current workspace to a file.
 * \param	fn	The filename to save the file to.
 * \return	An integer error code. -1 = error opening file. 0 = success.	*/
int ws_save_file(char *fn)
{
	FILE *ws;
	char *qryTxt;
	char *newTxt;
	int qryNum;
	char *fname;
	char *url;
	ABuf *qry = buf_new(100);

	if ((fn != NULL) && (strlen(fn) > 0))
		fname = fn;
	else
		fname = cfg_get_default_wsname();

	ws = fopen(fname, "w");
	if (!ws) {
		perror("opening workspace file for writing");
		return -1;
	}

	if (gui_connection_ok()) {	/* Save the connection and queries	*/
		/* Save the connection URL	*/
		url = gui_get_connection_url();
		if ((url != NULL) && (strlen(url) > 0)) {
			qry = buf_strcpy(qry, "connect(\"");
			qry = buf_strcat(qry, url);
			qry = buf_strcat(qry, "\")\n");
			fprintf(ws, qry->b_dat);

			/* Now, run through the open queries */
			qryNum = 0;
			while ((qryTxt = query_get_text_no(qryNum))) {
				newTxt = strdup(qryTxt);
				if (newTxt) {
					/** \todo Change map_chars to properly escape carriadge returns.	*/
					map_chars(newTxt, '\n', ' ');
					qry = buf_strcpy(qry, "addQuery(\"");
					qry = buf_strcat(qry, newTxt);
					qry = buf_strcat(qry, "\")\n");
					fprintf(ws, qry->b_dat);
					g_free(newTxt);
				}
				newTxt = NULL;
				qryNum++;
			}
		}
	}

	/* Cleanup.	*/
	qry = buf_free(qry);
	fclose(ws);
	return 0;
}
